"""Kubetest wrapper for the Kubernetes ``Container`` API Object."""

import logging

from kubernetes import client

log = logging.getLogger('kubetest')


class Container:
    """Kubetest wrapper around a Kubernetes `Container`_ API Object.

    The actual ``kubernetes.client.V1Container`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `Container`_.

    This wrapper does **NOT** subclass the ``objects.ApiObject`` like other
    object wrappers because it is not intended to be created or
    managed from manifest file. It is merely meant to wrap the
    Container spec for a Pod to make Container-targeted actions
    easier.

    .. _Container:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#container-v1-core
    """

    def __init__(self, api_object, pod):
        self.obj = api_object
        self.pod = pod

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def get_restart_count(self):
        """Get the number of times the Container has been restarted.

        Returns:
            int: The number of times the Container has been restarted.
        """
        container_name = self.obj.name
        pod_status = self.pod.status()

        # If there are no container status, the container hasn't started
        # yet, so there cannot be any restarts.
        if pod_status.container_statuses is None:
            return 0

        for status in pod_status.container_statuses:
            if status.name == container_name:
                return status.restart_count

        raise RuntimeError(
            'Unable to determine container status for {}'.format(container_name)
        )

    def get_logs(self):
        """Get all the logs for the Container.

        Returns:
            str: The Container logs.
        """
        return client.CoreV1Api().read_namespaced_pod_log(
            name=self.pod.name,
            namespace=self.pod.namespace,
            container=self.obj.name,
        )

    def search_logs(self, *keyword):
        """Search for keywords/phrases in the Container's logs.

        Args:
            *keyword (str): Keywords to search for within the logs.

        Returns:
            bool: True if found; False otherwise.
        """
        logs = self.get_logs()

        for k in keyword:
            if logs.find(k) == -1:
                return False
        return True



"""Define test conditions for kubetest."""

import enum


class Policy(enum.Enum):
    """Condition checking policies.

    A Policy defines the behavior of how Conditions are checked.

      - **ONCE**: A condition only needs to be met once and the check
        will consider it met regardless of the state of any other
        conditions that may be checked alongside it. This is the
        default behavior.
      - **SIMULTANEOUS**: A condition needs to be met simultaneous to
        all other conditions that are being checked alongside it for
        the check to be successful.
    """

    ONCE = 1
    SIMULTANEOUS = 2


class Condition:
    """A Condition is a convenience wrapper around a function and its arguments
    which allows the function to be called at a later time.

    The function is called in the ``check`` method, which resolves the result to
    a boolean value, thus the condition function should return a boolean or
    something that ultimately resolves to a Truthy or Falsey value.

    Args:
        name (str): The name of the condition to make it easier to identify.
        fn: The condition function that will be checked.
        *args: Any arguments for the condition function.
        **kwargs: Any keyword arguments for the condition function.

    Attributes:
        name (str): The name of the Condition.
        fn (callable): The condition function that will be checked.
        args (tuple): Arguments for the checking function.
        kwargs (dict): Keyword arguments for the checking function.
        last_check(bool): Holds the state of the last condition check.

    Raises:
        ValueError: The given ``fn`` is not callable.
    """

    def __init__(self, name, fn, *args, **kwargs):
        if not callable(fn):
            raise ValueError('The Condition function must be callable')

        self.name = name
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

        # last check holds the state of the last check.
        self.last_check = False

    def __str__(self):
        return '<Condition (name: {}, met: {})>'.format(
            self.name, self.last_check
        )

    def __repr__(self):
        return self.__str__()

    def check(self):
        """Check that the condition was met.

        Returns:
            bool: True if the condition was met; False otherwise.
        """
        self.last_check = bool(self.fn(*self.args, **self.kwargs))
        return self.last_check


def check_all(*args):
    """Check all the given Conditions.

    Args:
        *args (Condition): The Conditions to check.

    Returns:
        bool: True if all checks pass; False otherwise.
    """
    return all([cond.check() for cond in args])


def check_and_sort(*args):
    """Check all the given Conditions and sort them into 'met' and
    'unmet' buckets.

    Args:
        *args (Condition): The Conditions to check.

    Returns:
        tuple[list[Condition], list[Condition]]: The met and unmet
        condition buckets (in that order).
    """
    met, unmet = [], []

    for c in args:
        if c.check():
            met.append(c)
        else:
            unmet.append(c)

    return met, unmet



"""Kubetest wrapper for the Kubernetes ``Namespace`` API Object."""

import logging

from kubernetes import client

from kubetest.objects import ApiObject

log = logging.getLogger('kubetest')


class Namespace(ApiObject):
    """Kubetest wrapper around a Kubernetes `Namespace`_ API Object.

    The actual ``kubernetes.client.V1Namespace`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `Namespace`_.

    .. _Namespace:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#namespace-v1-core
    """

    obj_type = client.V1Namespace

    api_clients = {
        'preferred': client.CoreV1Api,
        'v1': client.CoreV1Api,
    }

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    @classmethod
    def new(cls, name):
        """Create a new Namespace with object backing.

        Args:
            name (str): The name of the new Namespace.

        Returns:
            Namespace: A new Namespace instance.
        """
        return cls(client.V1Namespace(
            metadata=client.V1ObjectMeta(
                name=name
            )
        ))

    def create(self, name=None):
        """Create the Namespace under the given name.

        Args:
            name (str): The name to create the Namespace under. If the
                name is not provided, it will be assumed to already be
                in the underlying object spec. If it is not, namespace
                operations will fail.
        """
        if name is not None:
            self.name = name

        log.info('creating namespace "%s"', self.name)
        log.debug('namespace: %s', self.obj)

        self.obj = self.api_client.create_namespace(
            body=self.obj,
        )

    def delete(self, options=None):
        """Delete the Namespace.

        Args:
             options (client.V1DeleteOptions): Options for Namespace deletion.

        Returns:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting namespace "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('namespace: %s', self.obj)

        return self.api_client.delete_namespace(
            name=self.name,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes Namespace resource."""
        self.obj = self.api_client.read_namespace(
            name=self.name,
        )

    def is_ready(self):
        """Check if the Namespace is in the ready state.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        self.refresh()

        status = self.obj.status
        if status is None:
            return False

        return status.phase.lower() == 'active'



"""Kubetest base class for the Kubernetes API Object wrappers."""

import abc
import logging

from kubernetes.client.rest import ApiException

from kubetest import condition, utils
from kubetest.manifest import load_file

log = logging.getLogger('kubetest')


class ApiObject(abc.ABC):
    """ApiObject is the base class for many of the kubetest objects
    which wrap Kubernetes API objects.

    This base class provides common functionality and common object
    properties for all API wrappers. It also defines the following
    abstract methods which all subclasses must implement:

      - ``create``: create the resource on the cluster
      - ``delete``: remove the resource from the cluster
      - ``refresh``: refresh the underlying object model
      - ``is_ready``: check if the object is in the ready state

    Args:
         api_object: The underlying Kubernetes API object.

    Attributes:
        obj: The underlying Kubernetes API object.
    """

    # The Kubernetes API object type. Each subclass should
    # define its own obj_type.
    obj_type = None
    '''The default Kubernetes API object type for the class.
    Each subclass should define its own ``obj_type``.
    '''

    api_clients = None
    '''A mapping of all the supported api clients for the API
    object type. Various resources can have multiple versions,
    e.g. "apps/v1", "apps/v1beta1", etc. The preferred version
    for each resource type should be defined under the "preferred"
    key. The preferred API client will be used when the apiVersion
    is not specified for the resource.
    '''

    def __init__(self, api_object):
        # The underlying Kubernetes Api Object
        self.obj = api_object

        # The api client for the object. This will be determined
        # by the apiVersion of the object's manifest.
        self._api_client = None

    @property
    def version(self):
        """str: The API version of the Kubernetes object (`obj.apiVersion``)."""
        return self.obj.api_version

    @property
    def name(self):
        """str: The name of the Kubernetes object (``obj.metadata.name``)."""
        return self.obj.metadata.name

    @name.setter
    def name(self, name):
        """Set the name of the Kubernetes objects (``obj.metadata.name``)."""
        self.obj.metadata.name = name

    @property
    def namespace(self):
        """The namespace of the Kubernetes object (``obj.metadata.namespace``)."""
        return self.obj.metadata.namespace

    @namespace.setter
    def namespace(self, name):
        """Set the namespace of the object, if it hasn't already been set.

        Raises:
            AttributeError: The namespace has already been set.
        """
        if self.obj.metadata.namespace is None:
            self.obj.metadata.namespace = name
        else:
            raise AttributeError('Cannot set namespace - object already has a namespace')

    @property
    def api_client(self):
        """The API client for the Kubernetes object. This is determined
        by the ``apiVersion`` of the object configuration.

        Raises:
            ValueError: The API version is not supported.
        """
        if self._api_client is None:
            c = self.api_clients.get(self.version)
            # If we didn't find the client in the api_clients dict, use the
            # preferred version.
            if c is None:
                log.warning(
                    'unknown version ({}), falling back to preferred version'
                    .format(self.version)
                )
                c = self.api_clients.get('preferred')
                if c is None:
                    raise ValueError(
                        'unknown version specified and no preferred version '
                        'defined for resource ({})'.format(self.version)
                    )
            # If we did find it, initialize that client version.
            self._api_client = c()
        return self._api_client

    def wait_until_ready(self, timeout=None, interval=1, fail_on_api_error=False):
        """Wait until the resource is in the ready state.

        Args:
            timeout (int): The maximum time to wait, in seconds, for
                the resource to reach the ready state. If unspecified,
                this will wait indefinitely. If specified and the timeout
                is met or exceeded, a TimeoutError will be raised.
            interval (int|float): The time, in seconds, to wait before
                re-checking if the object is ready.
            fail_on_api_error (bool): Fail if an API error is raised. An
                API error can be raised for a number of reasons, such as
                'resource not found', which could be the case when a resource
                is just being started or restarted. When waiting for readiness
                we generally do not want to fail on these conditions.
                (default: False)

        Raises:
             TimeoutError: The specified timeout was exceeded.
        """
        ready_condition = condition.Condition(
            'api object ready',
            self.is_ready,
        )

        utils.wait_for_condition(
            condition=ready_condition,
            timeout=timeout,
            interval=interval,
            fail_on_api_error=fail_on_api_error,
        )

    def wait_until_deleted(self, timeout=None, interval=1):
        """Wait until the resource is deleted from the cluster.

        Args:
            timeout (int): The maximum time to wait, in seconds, for
                the resource to be deleted from the cluster. If
                unspecified, this will wait indefinitely. If specified
                and the timeout is met or exceeded, a TimeoutError will
                be raised.
            interval (int|float): The time, in seconds, to wait before
                re-checking if the object has been deleted.

        Raises:
            TimeoutError: The specified timeout was exceeded.
        """
        def deleted_fn():
            try:
                self.refresh()
            except ApiException as e:
                # If we can no longer find the deployment, it is deleted.
                # If we get any other exception, raise it.
                if e.status == 404 and e.reason == 'Not Found':
                    return True
                else:
                    log.error('error refreshing object state')
                    raise e
            else:
                # The object was still found, so it has not been deleted
                return False

        delete_condition = condition.Condition(
            'api object deleted',
            deleted_fn
        )

        utils.wait_for_condition(
            condition=delete_condition,
            timeout=timeout,
            interval=interval,
        )

    @classmethod
    def load(cls, path):
        """Load the Kubernetes resource from file.

        Generally, this is used to load the Kubernetes manifest files
        and parse them into their appropriate API Object type.

        Args:
            path (str): The path to the YAML config file (manifest)
                containing the configuration for the resource.

        Returns:
            ApiObject: The API object wrapper corresponding to the configuration
            loaded from manifest YAML file.
        """
        objs = load_file(path)
        if len(objs) != 1:
            raise ValueError(
                'Unable to load resource from file - multiple resources found '
                'in specified file.'
            )
        return cls(objs[0])

    @abc.abstractmethod
    def create(self, namespace=None):
        """Create the underlying Kubernetes resource in the cluster
        under the given namespace.

        Args:
            namespace (str): The namespace to create the resource under.
                If no namespace is provided, it will use the instance's
                namespace member, which is set when the object is created
                via the kubetest client. (optional)
        """

    @abc.abstractmethod
    def delete(self, options):
        """Delete the underlying Kubernetes resource from the cluster.

        This method expects the resource to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will need
        to be set manually.

        Args:
            options (client.V1DeleteOptions): Options for resource deletion.
        """

    @abc.abstractmethod
    def refresh(self):
        """Refresh the local state (``obj``) of the underlying Kubernetes resource."""

    @abc.abstractmethod
    def is_ready(self):
        """Check if the resource is in the ready state.

        It is up to the wrapper subclass to define what "ready" means for
        that particular resource.

        Returns:
            bool: True if in the ready state; False otherwise.
        """



"""The test client for managing Kubernetes resources within test cases.

An instance of the ``TestClient`` defined in this module is automatically
created for each test case that uses the ``kube`` fixture. The ``kube``
fixture provides the ``TestClient`` instance to the test case.
"""

import logging

from kubernetes import client

from kubetest import objects, utils
from kubetest.condition import Condition, Policy, check_and_sort

log = logging.getLogger('kubetest')


class TestClient:
    """Test client for managing Kubernetes resources for a test case.

    The ``namespace`` for the TestClient will be automatically generated
    and provided to the TestClient during the test setup process.

    Args:
        namespace (str): The namespace associated with the test client.
            Each test case will have its own namespace assigned.
    """

    def __init__(self, namespace):
        self.namespace = namespace
        self.pre_registered = []

    # ****** Manifest Loaders ******

    @staticmethod
    def load_clusterrolebinding(path):
        """Load a manifest YAML into a ClusterRoleBinding object.

        Args:
            path (str): The path to the ClusterRoleBinding manifest.

        Returns:
            objects.ClusterRoleBinding: The ClusterRoleBinding for the
            specified manifest.
        """
        log.info('loading clusterrolebinding from path: %s', path)
        clusterrolebinding = objects.ClusterRoleBinding.load(path)
        return clusterrolebinding

    def load_configmap(self, path, set_namespace=True):
        """Load a manifest YAML into a ConfigMap object.

        By default, this will augment the ConfigMap object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the ConfigMap manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the ConfigMap namespace.

        Returns:
            objects.ConfigMap: The ConfigMap for the specified manifest.
        """
        log.info('loading configmap from path: %s', path)
        configmap = objects.ConfigMap.load(path)
        if set_namespace:
            configmap.namespace = self.namespace
        return configmap

    def load_deployment(self, path, set_namespace=True):
        """Load a manifest YAML into a Deployment object.

        By default, this will augment the Deployment object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the Deployment manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the Deployment namespace.

        Returns:
            objects.Deployment: The Deployment for the specified manifest.
        """
        log.info('loading deployment from path: %s', path)
        deployment = objects.Deployment.load(path)
        if set_namespace:
            deployment.namespace = self.namespace
        return deployment

    def load_statefulset(self, path, set_namespace=True):
        """Load a manifest YAML into a StatefulSet object.

        By default, this will augment the StatefulSet object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the StatefulSet manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the StatefulSet namespace.

        Returns:
            objects.StatefulSet: The StatefulSet for the specified manifest.
        """
        log.info('loading statefulset from path: %s', path)
        statefulset = objects.StatefulSet.load(path)
        if set_namespace:
            statefulset.namespace = self.namespace
        return statefulset

    def load_daemonset(self, path, set_namespace=True):
        """Load a manifest YAML into a DaemonSet object.

        By default, this will augment the DaemonSet object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the DaemonSet manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the DaemonSet namespace.

        Returns:
            objects.DaemonSet: The DaemonSet for the specified manifest.
        """
        log.info('loading daemonset from path: %s', path)
        daemonset = objects.DaemonSet.load(path)
        if set_namespace:
            daemonset.namespace = self.namespace
        return daemonset

    def load_pod(self, path, set_namespace=True):
        """Load a manifest YAML into a Pod object.

        By default, this will augment the Pod object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the Pod manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the Pod namespace.

        Returns:
            objects.Pod: The Pod for the specified manifest.
        """
        log.info('loading pod from path: %s', path)
        pod = objects.Pod.load(path)
        if set_namespace:
            pod.namespace = self.namespace
        return pod

    def load_rolebinding(self, path, set_namespace=True):
        """Load a manifest YAML into a RoleBinding object.

        By default, this will augment the RoleBinding object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the RoleBinding manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the RoleBinding namespace.

        Returns:
            objects.RoleBinding: The RoleBinding for the specified manifest.
        """
        log.info('loading rolebinding from path: %s', path)
        rolebinding = objects.RoleBinding.load(path)
        if set_namespace:
            rolebinding.namespace = self.namespace
        return rolebinding

    def load_secret(self, path, set_namespace=True):
        """Load a manifest YAML into a Secret object.

        By default, this will augment the Secret object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the Secret manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the Secret namespace.

        Returns:
            objects.Secret: The Secret for the specified manifest.
        """
        log.info('loading secret from path: %s', path)
        secret = objects.Secret.load(path)
        if set_namespace:
            secret.namespace = self.namespace
        return secret

    def load_service(self, path, set_namespace=True):
        """Load a manifest YAML into a Service object.

        By default, this will augment the Service object with
        the generated test case namespace. This behavior can be
        disabled with the ``set_namespace`` flag.

        Args:
            path (str): The path to the Service manifest.
            set_namespace (bool): Enable/disable the automatic
                augmentation of the Service namespace.

        Returns:
            objects.Service: The Service for the specified manifest.
        """
        log.info('loading service from path: %s', path)
        service = objects.Service.load(path)
        if set_namespace:
            service.namespace = self.namespace
        return service

    # ****** Generic Helpers on ApiObjects ******

    def create(self, obj):
        """Create the provided ApiObject on the Kubernetes cluster.

        If the object does not already have a namespace assigned to it,
        the client's generated test case namespace will be used.

        Args:
            obj (objects.ApiObject): A kubetest API Object wrapper.
        """
        if obj.namespace is None:
            obj.namespace = self.namespace

        obj.create()

    def delete(self, obj, options=None):
        """Delete the provided ApiObject from the Kubernetes cluster.

        If the object does not already have a namespace assigned to it,
        the client's generated test case namespace will be used.

        Args:
            obj (objects.ApiObject): A kubetest API Object wrapper.
            options (client.V1DeleteOptions): Additional options for
                deleting the resource from the cluster.
        """
        if obj.namespace is None:
            obj.namespace = self.namespace
        if options is None:
            options = client.V1DeleteOptions()

        obj.delete(options=options)

    @staticmethod
    def refresh(obj):
        """Refresh the underlying Kubernetes resource status and state.

        Args:
            obj (objects.ApiObject): A kubetest API Object wrapper.
        """
        obj.refresh()

    # ****** General Helpers ******

    def get_namespaces(self, fields=None, labels=None):
        """Get Namespaces from the cluster.

        Args:
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Namespaces to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Namespaces to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.Namespace]: A dictionary where the key is
            the Namespace name and the value is the Nameself itself.
        """
        selectors = utils.selector_kwargs(fields, labels)

        namespace_list = client.CoreV1Api().list_namespace(
            **selectors,
        )

        namespaces = {}
        for obj in namespace_list.items:
            namespace = objects.Namespace(obj)
            namespaces[namespace.name] = namespace

        return namespaces

    def get_deployments(self, namespace=None, fields=None, labels=None):
        """Get Deployments from the cluster.

        Args:
            namespace (str): The namespace to get the Deployments from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Deployments to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Deployments to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.Deployment]: A dictionary where the key is
            the Deployment name and the value is the Deployment itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        deployment_list = client.AppsV1Api().list_namespaced_deployment(
            namespace=namespace,
            **selectors,
        )

        deployments = {}
        for obj in deployment_list.items:
            deployment = objects.Deployment(obj)
            deployments[deployment.name] = deployment

        return deployments

    def get_statefulsets(self, namespace=None, fields=None, labels=None):
        """Get StatefulSets from the cluster.

        Args:
            namespace (str): The namespace to get the StatefulSets from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of StatefulSets to only those which
                match these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of StatefulSets to only those which
                match these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.StatefulSet]: A dictionary where the key is
            the StatefulSet name and the value is the StatefulSet itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        statefulset_list = client.AppsV1Api().list_namespaced_stateful_set(
            namespace=namespace,
            **selectors,
        )

        statefulsets = {}
        for obj in statefulset_list.items:
            statefulset = objects.StatefulSet(obj)
            statefulsets[statefulset.name] = statefulset

        return statefulsets

    def get_daemonsets(self, namespace=None, fields=None, labels=None):
        """Get DaemonSets from the cluster.

        Args:
            namespace (str): The namespace to get the DaemonSets from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of DaemonSets to only those which
                match these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of DaemonSets to only those which
                match these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.DaemonSet]: A dictionary where the key is
            the DaemonSet name and the value is the DaemonSet itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        daemonset_list = client.AppsV1Api().list_namespaced_daemon_set(
            namespace=namespace,
            **selectors,
        )

        daemonsets = {}
        for obj in daemonset_list.items:
            daemonset = objects.DaemonSet(obj)
            daemonsets[daemonset.name] = daemonset

        return daemonsets

    def get_endpoints(self, namespace=None, fields=None, labels=None):
        """Get Endpoints from the cluster.

        Args:
            namespace (str): The namespace to get the Endpoints from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Endpoints to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Endpoints to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.Endpoints]: A dictionary where the key is
            the Endpoint name and the value is the Endpoint itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        endpoints_list = client.CoreV1Api().list_namespaced_endpoints(
            namespace=namespace,
            **selectors,
        )

        endpoints = {}
        for obj in endpoints_list.items:
            endpoint = objects.Endpoints(obj)
            endpoints[endpoint.name] = endpoint

        return endpoints

    def get_secrets(self, namespace=None, fields=None, labels=None):
        """Get Secrets from the cluster.

        Args:
            namespace (str): The namespace to get the Secrets from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Secrets to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Secrets to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.Secret]: A dictionary where the key is
            the Secret name and the value is the Secret itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        secret_list = client.CoreV1Api().list_namespaced_secret(
            namespace=namespace,
            **selectors,
        )

        secrets = {}
        for obj in secret_list.items:
            secret = objects.Secret(obj)
            secrets[secret.name] = secret

        return secrets

    def get_configmaps(self, namespace=None, fields=None, labels=None):
        """Get ConfigMaps from the cluster.

        Args:
            namespace (str): The namespace to get the ConfigMaps from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of ConfigMaps to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of ConfigMaps to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.ConfigMap]: A dictionary where the key is the
            ConfigMap name and the value is the ConfigMap itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        configmap_list = client.CoreV1Api().list_namespaced_config_map(
            namespace=namespace,
            **selectors,
        )

        configmaps = {}
        for obj in configmap_list.items:
            cm = objects.ConfigMap(obj)
            configmaps[cm.name] = cm

        return configmaps

    def get_pods(self, namespace=None, fields=None, labels=None):
        """Get Pods from the cluster.

        Args:
            namespace (str): The namespace to get the Pods from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Pods to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Pods to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.Pod]: A dictionary where the key is the Pod
            name and the value is the Pod itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        pod_list = client.CoreV1Api().list_namespaced_pod(
            namespace=namespace,
            **selectors,
        )

        pods = {}
        for obj in pod_list.items:
            pod = objects.Pod(obj)
            pods[pod.name] = pod

        return pods

    def get_services(self, namespace=None, fields=None, labels=None):
        """Get Services under the test case namespace.

        Args:
            namespace (str): The namespace to get the Services from. If not
                specified, it will use the auto-generated test case namespace
                by default.
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Services to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Services to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.Service]: A dictionary where the key is the
            Service name and the value is the Service itself.
        """
        if namespace is None:
            namespace = self.namespace

        selectors = utils.selector_kwargs(fields, labels)

        service_list = client.CoreV1Api().list_namespaced_service(
            namespace=namespace,
            **selectors,
        )

        services = {}
        for obj in service_list.items:
            service = objects.Service(obj)
            services[service.name] = service

        return services

    @staticmethod
    def get_nodes(fields=None, labels=None):
        """Get the Nodes that make up the cluster.

        Args:
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Nodes to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Nodes to only those which match
                these label selectors. By default, no restricting is done.

        Returns:
            dict[str, objects.Node]: A dictionary where the key is the Node
            name and the value is the Node itself.
        """
        selectors = utils.selector_kwargs(fields, labels)

        node_list = client.CoreV1Api().list_node(
            **selectors,
        )

        nodes = {}
        for obj in node_list.items:
            node = objects.Node(obj)
            nodes[node.name] = node

        return nodes

    def get_events(self, fields=None, labels=None, all_namespaces=False):
        """Get the latest Events that occurred in the cluster.

        Args:
            fields (dict[str, str]): A dictionary of fields used to restrict
                the returned collection of Events to only those which match
                these field selectors. By default, no restricting is done.
            labels (dict[str, str]): A dictionary of labels used to restrict
                the returned collection of Events to only those which match
                these label selectors. By default, no restricting is done.
            all_namespaces (bool): If True, get the events across all
                namespaces.

        Returns:
            dict[str, objects.Event]: A dictionary where the key is the Event
            name and the value is the Event itself.
        """
        selectors = utils.selector_kwargs(fields, labels)

        if all_namespaces:
            event_list = client.CoreV1Api().list_event_for_all_namespaces(
                **selectors
            )
        else:
            event_list = client.CoreV1Api().list_namespaced_event(
                namespace=self.namespace,
                **selectors
            )

        events = {}
        for obj in event_list.items:
            event = objects.Event(obj)
            events[event.name] = event

        return events

    # ****** Test Helpers ******

    @staticmethod
    def wait_for_conditions(
            *args, timeout=None, interval=1, policy=Policy.ONCE, fail_on_api_error=True
    ):
        """Wait for all of the provided Conditions to be met.

        All Conditions must be met for this to unblock. If no Conditions are
        provided, this method will do nothing.

        Args:
            *args (Condition): Conditions to check.
            timeout (int): The maximum time to wait, in seconds, for the
                provided Conditions to be met. If all of the Conditions are
                not met within the given timeout, this will raise a TimeoutError.
                By default, there is no timeout so this will wait indefinitely.
            interval (float|int): The time, in seconds, to sleep before
                re-evaluating the conditions. Default: 1s
            policy (condition.Policy): The condition checking policy that defines
                the checking behavior. Default: ONCE
            fail_on_api_error (bool): Fail the condition checks if a Kubernetes
                API error is incurred. An API error can be raised for a number
                of reasons, including a Pod being restarted and temporarily
                unavailable. Disabling this will cause those errors to be
                ignored, allowing the check to continue until timeout or
                resolution. (default: True).

        Raises:
            TimeoutError: The Conditions were not met within the specified
                timeout period.
            ValueError: Not all arguments are a Condition.
        """
        # If no Conditions were given, there is nothing to do.
        if not args:
            return

        # If something was given, make sure they are all Conditions
        if not all(map(lambda c: isinstance(c, Condition), args)):
            raise ValueError('All arguments must be a Condition')

        # make a copy of the conditions
        to_check = list(args)

        def condition_checker(conditions):
            # check that the conditions were met according to the
            # condition checking policy
            met, unmet = check_and_sort(*conditions)
            if policy == Policy.ONCE:
                log.info('check met: %s', met)
                conditions[:] = unmet
                return len(unmet) == 0

            elif policy == Policy.SIMULTANEOUS:
                return len(unmet) == 0 and len(met) == len(args)

            else:
                raise ValueError(
                    'Invalid condition policy specified: {}'.format(policy)
                )

        wait_condition = Condition(
            'wait for conditions',
            condition_checker,
            to_check,
        )

        try:
            utils.wait_for_condition(
                condition=wait_condition,
                timeout=timeout,
                interval=interval,
                fail_on_api_error=fail_on_api_error,
            )
        except TimeoutError:
            # If we time out here, we want to show all the conditions
            # that we weren't able to resolve in the error message, not
            # the 'wait for conditions' wrapper.
            raise TimeoutError(
                'timed out wile waiting for conditions to be met: {}'.format(to_check)
            )

    def wait_for_ready_nodes(self, count, timeout=None, interval=1):
        """Wait until there are at least ``count`` number of nodes available
        in the cluster.

        Notes:
            This should only be used for clusters that auto-scale the
            nodes. This will not create/delete nodes on its own.

        Args:
            count (int): The number of nodes to wait for.
            timeout (int): The maximum time to wait, in seconds.
            interval (int|float): The time, in seconds, to sleep before
                re-checking the number of nodes.
        """
        def node_count_match(node_count):
            nodes = self.get_nodes()
            return [n.is_ready() for n in nodes.values()].count(True) >= node_count

        wait_condition = Condition(
            'wait for {} nodes'.format(count),
            node_count_match,
            count,
        )

        utils.wait_for_condition(
            condition=wait_condition,
            timeout=timeout,
            interval=interval,
        )

    def wait_for_registered(self, timeout=None, interval=1):
        """Wait for all of the pre-registered objects to be ready on the cluster.

        An object is pre-registered with the test client if it is specified
        to the test via the ``applymanifests`` pytest marker. The marker will load
        the manifest and add the object to the cluster, and register it with
        the test client. This method waits until all such loaded manifest objects
        are in the ready state simultaneously.

        Args:
            timeout (int): The maximum time to wait, in seconds.
            interval (int|float): The time, in seconds, to sleep before
                re-checking the ready state for pre-registered objects.
        """
        def check_registered():
            for obj in self.pre_registered:
                if not obj.is_ready():
                    return False
            return True

        wait_condition = Condition(
            'wait for pre-registered objects to be ready',
            check_registered,
        )

        utils.wait_for_condition(
            condition=wait_condition,
            timeout=timeout,
            interval=interval,
        )

    @staticmethod
    def wait_until_created(obj, timeout=None, interval=1):
        """Wait until the specified object has been created.

        Here, creation is judged on whether or not refreshing the
        object (e.g. getting it) returns an object (created) or
        an error (not yet created).

        Args:
            obj (objects.ApiObject): The ApiObject to wait on.
            timeout (int): The maximum time to wait, in seconds.
            interval (int|float): The time, in seconds, to sleep before
                re-checking the created state of the object.
        """
        def check_ready(api_obj):
            try:
                api_obj.refresh()
            except:  # noqa
                return False
            return True

        wait_condition = Condition(
            'wait for {}:{} to be created'.format(type(obj).__name__, obj.name),
            check_ready,
            obj,
        )

        utils.wait_for_condition(
            condition=wait_condition,
            timeout=timeout,
            interval=interval
        )



"""Kubetest wrapper for the Kubernetes ``Service`` API Object."""

import logging

from kubernetes import client

from .api_object import ApiObject

log = logging.getLogger('kubetest')


class Service(ApiObject):
    """Kubetest wrapper around a Kubernetes `Service`_ API Object.

    The actual ``kubernetes.client.V1Service`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `Service`_.

    .. _Service:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#service-v1-core
    """

    obj_type = client.V1Service

    api_clients = {
        'preferred': client.CoreV1Api,
        'v1': client.CoreV1Api,
    }

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def create(self, namespace=None):
        """Create the Service under the given namespace.

        Args:
            namespace (str): The namespace to create the Service under.
                If the Service was loaded via the kubetest client, the
                namespace will already be set, so it is not needed here.
                Otherwise, the namespace will need to be provided.
        """
        if namespace is None:
            namespace = self.namespace

        log.info('creating service "%s" in namespace "%s"', self.name, self.namespace)
        log.debug('service: %s', self.obj)

        self.obj = self.api_client.create_namespaced_service(
            namespace=namespace,
            body=self.obj,
        )

    def delete(self, options):
        """Delete the Service.

        This method expects the Service to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will need
        to be set manually.

        Args:
            options (client.V1DeleteOptions): Options for Service deletion.

        Returns:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting service "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('service: %s', self.obj)

        return self.api_client.delete_namespaced_service(
            name=self.name,
            namespace=self.namespace,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes Service resource."""
        self.obj = self.api_client.read_namespaced_service(
            name=self.name,
            namespace=self.namespace,
        )

    def is_ready(self):
        """Check if the Service is in the ready state.

        The readiness state is not clearly available from the Service
        status, so to see whether or not the Service is ready this
        will check whether the endpoints of the Service are ready.

        This comes with the caveat that in order for a Service to
        have endpoints, there needs to be some backend hooked up to it.
        If there is no backend, the Service will never have endpoints,
        so this will never resolve to True.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        self.refresh()

        # check the status. if there is no status, the service is
        # definitely not ready.
        if self.obj.status is None:
            return False

        endpoints = self.get_endpoints()

        # if the Service has no endpoints, its not ready.
        if len(endpoints) == 0:
            return False

        # get the service endpoints and check that they are all ready.
        for endpoint in endpoints:
            # if we have an endpoint, but there are no subsets, we
            # consider the endpoint to be not ready.
            if endpoint.subsets is None:
                return False

            for subset in endpoint.subsets:
                # if the endpoint has no addresses setup yet, its not ready
                if subset.addresses is None or len(subset.addresses) == 0:
                    return False

                # if there are still addresses that are not ready, the
                # service is not ready
                not_ready = subset.not_ready_addresses
                if not_ready is not None and len(not_ready) > 0:
                    return False

        # if we got here, then all endpoints are ready, so the service
        # must also be ready
        return True

    def status(self):
        """Get the status of the Service.

        Returns:
            client.V1ServiceStatus: The status of the Service.
        """
        log.info('checking status of service "%s"', self.name)
        # first, refresh the service state to ensure the latest status
        self.refresh()

        # return the status from the service
        return self.obj.status

    def get_endpoints(self):
        """Get the endpoints for the Service.

        This can be useful for checking internal IP addresses used
        in containers, e.g. for container auto-discovery.

        Returns:
            list[client.V1Endpoints]: A list of endpoints associated
            with the Service.
        """
        log.info('getting endpoints for service "%s"', self.name)
        endpoints = self.api_client.list_namespaced_endpoints(
            namespace=self.namespace,
        )

        svc_endpoints = []
        for endpoint in endpoints.items:
            # filter to include only the endpoints with the same
            # name as the service.
            if endpoint.metadata.name == self.name:
                svc_endpoints.append(endpoint)

        log.debug('endpoints: %s', svc_endpoints)
        return svc_endpoints

    def proxy_http_get(self, path):
        """Issue a GET request to proxy of a Service.

        Args:
            path (str): The URI path for the request.

        Returns:
            str: The response data
        """
        return client.CoreV1Api().connect_get_namespaced_service_proxy_with_path(
            name="{}:{}".format(self.name, self.obj.spec.ports[0].port),
            namespace=self.namespace,
            path=path,
        )



"""Kubetest manager for test client instances and namespace management."""

import logging

import kubernetes

from kubetest import client, objects, utils

log = logging.getLogger('kubetest')


class ObjectManager:
    """ObjectManager is a convenience class used to manage Kubernetes API
    objects that are registered with a test case.

    The core usage of the ObjectManager is to sort each of the registered
    objects into different buckets by type. An "apply order" is also
    defined here so we can get the bucketed objects in the order that they
    should be applied onto the cluster.

    This manager will only be used for API objects loaded from manifests
    that are specified by the `pytest.mark.applymanifests` marker on a
    test case by default.
    """

    # The ApiObject buckets in the order that that should be
    # applied when creating them on the cluster.
    ordered_buckets = [
        'namespace',
        'rolebinding',
        'clusterrolebinding',
        'secret',
        'service',
        'configmap',
        'daemonset',
        'statefulset',
        'deployment',
        'pod',
    ]

    def __init__(self):
        # Add attributes with the bucket names to the instance.
        # By default they will all be empty lists. While this could
        # be done in a more readable way by explicitly setting the
        # buckets as members, e.g. self.pod = [], but by tying the
        # creation to the ordered_buckets list (which the other
        # instance methods use), adding and removing buckets means
        # only adding to the list, not updating in numerous locations.
        for bucket in self.ordered_buckets:
            self.__setattr__(bucket, [])

    def add(self, *args):
        """Add API objects to the object manager.

        This method will take in any number of ApiObjects and sort
        them into the correct buckets. It will not check for duplicates.
        If a non-ApiObject is passed in, an error will be raised.

        Args:
            *args (objects.ApiObject): Any subclass of the kubetest
                ApiObject wrapping a Kubernetes API object.

        Raises:
            ValueError: One or more arguments passed to the function
                are not ApiObject subclasses.
        """
        for arg in args:
            if not isinstance(arg, objects.ApiObject):
                raise ValueError(
                    'Only ApiObject instances can be added to the ObjectManager, '
                    'but was given: {}'.format(arg)
                )

            # Get the type name of the ApiObject wrapper and lower case it,
            # e.g. ClusterRoleBinding -> clusterrolebinding. This will be
            # used to compare to the buckets.
            name = type(arg).__name__.lower()

            # Check if we have a bucket for the name, if so, add it to the
            # bucket. If not, raise an error.
            if name in self.ordered_buckets:
                self.__getattribute__(name).append(arg)
            else:
                raise ValueError(
                    'Unable to determine bucket for ApiObject: {}'.format(arg)
                )

    def get_objects_in_apply_order(self):
        """Get all of the managed objects in the order that they
        should be applied onto the cluster.

        Within the buckets themselves, API objects are not sorted.
        This function only yields the buckets in the correct order.

        Each of the buckets corresponds to an ApiObject wrapper that
        is supported by kubetest. As more ApiObject wrappers are added,
        the buckets here should be updated to reflect that.

        The bucket order in which objects are yielded are:
          - Namespace
          - RoleBinding
          - ClusterRoleBinding
          - Secret
          - Service
          - ConfigMap
          - DaemonSet
          - StatefulSet
          - Deployment
          - Pod

        Yields:
            ApiObject: The kubetest ApiObject wrapper to be created
                on the cluster.
        """
        for bucket in self.ordered_buckets:
            for obj in self.__getattribute__(bucket):
                yield obj


class TestMeta:
    """TestMeta holds information associated with a single test node.

    Args:
        name (str): The name of the test.
        node_id (str): The id of the test node.
    """

    def __init__(self, name, node_id, namespace_create, namespace_name):
        self.name = name
        self.node_id = node_id

        if namespace_name is None:
            self.ns = utils.new_namespace(name)
        else:
            self.ns = namespace_name

        # Flag to designate whether pytest setup (not TestMeta.setup) failed.
        # If pytest setup failed for the case, we can skip teardown, as there
        # will be nothing on the cluster itself.
        self._pt_setup_failed = False

        # lazy load
        self._client = None
        self._namespace = None

        self.namespace_create = namespace_create
        self.rolebindings = []
        self.clusterrolebindings = []

        self.test_objects = ObjectManager()

    @property
    def client(self):
        """Get the TestClient for the test case."""
        if self._client is None:
            self._client = client.TestClient(self.ns)
        return self._client

    @property
    def namespace(self):
        """Get the Namespace API Object associated with the test case."""
        if self._namespace is None:
            self._namespace = objects.Namespace.new(self.ns)
        return self._namespace

    def setup(self):
        """Setup the cluster state for the test case.

        This performs all actions needed in order for the test client to be
        ready to use by a test case.
        """
        # create the test case namespace
        if self.namespace_create:
            self.namespace.create()

        # if there are any role bindings, create them.
        for rb in self.rolebindings:
            self.client.create(rb)

        # if there are any cluster role bindings, create them.
        for crb in self.clusterrolebindings:
            self.client.create(crb)

        # if any objects were registered with the test case via the
        # `applymanifests` marker, register them to the test client
        # and add them to the cluster now
        for obj in self.test_objects.get_objects_in_apply_order():
            self.client.create(obj)
            self.client.wait_until_created(obj, timeout=10, interval=0.5)
            self.client.pre_registered.append(obj)

    def teardown(self):
        """Clean up the cluster state for the given test case.

        This performs all actions needed in order to clean up the state that
        was previously set up for the test client in `setup`.
        """
        # Only perform teardown if pytest setup was successful and there is a
        # possibility of things existing on the cluster.
        if self._pt_setup_failed:
            log.info('pytest setup failed for {}: not running test case teardown'.format(
                self.name)
            )
            return

        # Delete the test case namespace if we've created it.
        # This will also delete anything in the namespace, which includes RoleBindings.
        if self._namespace and self.namespace_create:
            self.namespace.delete()

        # ClusterRoleBindings are not bound to a namespace, so we will need
        # to delete them ourselves.
        for crb in self.clusterrolebindings:
            self.client.delete(crb)

    def yield_container_logs(self, tail_lines=None):
        """Yield the container logs for the test case.

        These logs will be printed out if the test was in error to provide
        more context and make it easier to debug the issue.

        Args:
            tail_lines (int): The number of container log lines to print.

        Yields:
            str: Logs for the running containers on the cluster.
        """
        try:
            # prior to tearing down the namespace and cleaning up all of the
            # objects in the namespace, get the logs for the containers in the
            # namespace.
            pods_list = kubernetes.client.CoreV1Api().list_namespaced_pod(
                namespace=self.ns
            )
        except Exception as e:
            log.warning(
                'Unable to get pods for namespace "%s" to cache logs (%s)',
                self.ns, e
            )
            return

        log_kwargs = {}
        if tail_lines is not None and tail_lines > 0:
            log_kwargs['tail_lines'] = tail_lines

        for pod in pods_list.items:
            for container in pod.spec.containers:
                pod_name = pod.metadata.name
                pod_ns = pod.metadata.namespace
                container_name = container.name
                try:
                    logs = kubernetes.client.CoreV1Api().read_namespaced_pod_log(
                        name=pod_name,
                        namespace=pod_ns,
                        container=container_name,
                        **log_kwargs,
                    )
                except Exception as e:
                    log.warning(
                        'Unable to cache logs for %s::%s (%s)',
                        pod_name, container_name, e
                    )
                    continue

                if logs != '':
                    _id = '=== {} -> {}::{} ==='.format(
                        self.node_id, pod_name, container_name
                    )
                    border = '=' * len(_id)
                    yield '\n'.join([border, _id, border, logs, '\n'])
        return

    def register_rolebindings(self, *rolebindings):
        """Register a RoleBinding requirement with the test case.

        Args:
            *rolebindings (RoleBinding): The RoleBindings that are needed for
                the test case.
        """
        self.rolebindings.extend(rolebindings)

    def register_clusterrolebindings(self, *clusterrolebindings):
        """Register a ClusterRoleBinding requirement with the test case.

        Args:
            *clusterrolebindings (ClusterRoleBinding): The ClusterRoleBindings
                that are needed for the test case.
        """
        self.clusterrolebindings.extend(clusterrolebindings)

    def register_objects(self, api_objects):
        """Register the provided objects with the test case.

        These objects will be registered to the test client and applied to
        the namespace on test setup.

        Args:
            api_objects (list): The wrapped Kubernetes API objects to create
                on the cluster.
        """
        self.test_objects.add(*api_objects)


class KubetestManager:
    """The manager for kubetest state.

    The KubetestManager is in charge of providing test clients for the tests
    that request them and mediating the namespace management corresponding to
    those test clients.
    """

    def __init__(self):

        # A dictionary mapping test node IDs to their corresponding TestMeta
        # object. Each test node will have a TestMeta created when the client
        # is created for the test node.
        self.nodes = {}

    def new_test(self, node_id, test_name, namespace_create, namespace_name):
        """Create a new TestMeta for a test case.

        This will be called by the test setup hook in order to create a new
        TestMeta for the manager to keep track of.

        Args:
            node_id (str): The id of the test node.
            test_name (str): The name of the test.

        Returns:
            TestMeta: The newly created TestMeta for the test case.
        """
        log.info('creating test meta for %s', node_id)
        meta = TestMeta(
            node_id=node_id,
            name=test_name,
            namespace_create=namespace_create,
            namespace_name=namespace_name
        )

        self.nodes[node_id] = meta
        return meta

    def get_test(self, node_id):
        """Get the test metadata for the specified test node.

        Args:
            node_id (str): The id of the test node.

        Returns:
            TestMeta: The test metadata for the given node.
            None: No test metadata was found for the given node.
        """
        return self.nodes.get(node_id)

    def teardown(self, node_id):
        """Tear down the test case.

        This is effectively a wrapper around the `teardown` method of the
        test client. It will also remove the test client from the manager.

        Test client teardown will delete the test client's namespace from
        the cluster. Deleting a namespace will delete all the things in the
        namespace (e.g. API objects bound to the namespace).

        Args:
            node_id (str): The id of the test node.
        """
        test_case = self.nodes.get(node_id)
        if test_case is not None:
            test_case.teardown()
            del self.nodes[node_id]



"""A pytest plugin which helps with integration testing for Kubernetes deployments.

This plugin helps with the managing of the Kubernetes cluster and provides
useful test fixtures and functionality in order to interact with and test
the state of the cluster.
"""

import logging
import os
import warnings

import kubernetes
import pytest
import urllib3

from kubetest import errors, markers
from kubetest.manager import KubetestManager

GOOGLE_APPLICATION_CREDENTIALS = 'GOOGLE_APPLICATION_CREDENTIALS'

log = logging.getLogger('kubetest')

# A global instance of the KubetestManager. This will be used by various
# pytest hooks and fixtures in order to create and manage the TestClient
# instances. The manager is in charge of creating new TestClients for each
# test that requests one via the `k8s` fixture, and to manage the lifecycle
# of the test namespace.
manager = KubetestManager()


# ********** pytest hooks **********

def pytest_addoption(parser):
    """Add options to pytest to configure kubetest.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_addoption
    """

    group = parser.getgroup('kubetest', 'kubernetes integration test support')
    group.addoption(
        '--kube-config',
        action='store',
        metavar='path',
        default=None,
        help=(
            'the kubernetes config for kubetest; this is required for '
            'resources to be installed on the cluster'
        )
    )
    group.addoption(
        '--kube-context',
        action='store',
        metavar='context',
        default=None,
        help='the name of the kubernetes config context to use',
    )
    group.addoption(
        '--kube-disable',
        action='store_true',
        default=False,
        help='[DEPRECATED] disable automatic configuration with the kubeconfig file'
    )

    # FIXME (etd) - this was an attempt to fix occasional permissions errors
    # (https://github.com/vapor-ware/kubetest/issues/11) but in doing so, it looks
    # like I hosed my permissions, so I'm just commenting all of this out for now...
    # group.addoption(
    #     '--google-application-credentials',
    #     action='store',
    #     metavar='path',
    #     default=None,
    #     help='google application credentials for GKE clusters'
    # )

    group.addoption(
        '--kube-log-level',
        action='store',
        default='warning',
        help='log level for the kubetest logger'
    )

    group.addoption(
        '--kube-error-log-lines',
        action='store',
        default=50,
        help='set the number of lines to tail from container logs on error. '
             'to show all lines, set this to -1.'
    )

    group.addoption(
        '--suppress-insecure-request',
        action='store',
        default=False,
        help='suppress the urllib3 InsecureRequestWarning. This is useful if testing '
             'against a cluster without HTTPS set up.'
    )


def pytest_report_header(config):
    """Augment the pytest report header with kubetest info.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_report_header
    """
    config_file = config.getoption('kube_config')
    if config_file is None:
        config_file = 'default'

    context = config.getoption('kube_context')
    if context is None:
        context = 'current context'

    return [
        'kubetest config file: {}'.format(config_file),
        'kubetest context: {}'.format(context),
    ]


def pytest_configure(config):
    """Configure pytest with kubetest additions.

    This registers the kubetest markers with pytest.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_configure
    """
    # Register kubetest markers
    markers.register(config)

    # Disable warnings, if configured to do so.
    if config.getoption('suppress_insecure_request'):
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def pytest_sessionstart(session):
    """Configure kubetest for the test session.

    Kubetest setup happens at session start (pytest_sessionstart) rather
    than on configuration (pytest_configure) so that we only have the
    expectation of a cluster config and available cluster when there are
    actually tests available. For example, if we are simply calling any of

        pytest --help
        pytest --markers
        pytest --fixtures

    we do not want to configure kubetest and force the expectation of
    cluster availability.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_sessionstart
    """
    # Setup the kubetest logger
    log_level = session.config.getoption('kube_log_level')
    level = logging._nameToLevel.get(log_level.upper(), logging.WARNING)
    logger = logging.getLogger('kubetest')
    logger.setLevel(level)

    # Check for configuration deprecations
    if session.config.getoption('kube_disable'):
        warnings.warn(
            '--kube-disable flag is deprecated (v0.2.0) and is no longer functional. '
            'To disable the plugin for a project, see: '
            'https://docs.pytest.org/en/latest/plugins.html',
        )

    # # Configure kubetest with the kubernetes config, if not disabled.
    # disabled = session.config.getoption('kube_disable')
    # if not disabled:
    #     # # Set the GOOGLE_APPLICATION_CREDENTIALS environment variable. For more, see:
    #     # # https://cloud.google.com/docs/authentication/production
    #     # gke_creds = config.getoption('google_application_credentials')
    #     # if gke_creds is not None:
    #     #     # If application credentials already exist, store them so they can be
    #     #     # reset after testing.
    #     #     old_creds = os.environ.get(GOOGLE_APPLICATION_CREDENTIALS)
    #     #     if old_creds:
    #     #         os.environ['OLD_' + GOOGLE_APPLICATION_CREDENTIALS] = old_creds
    #     #     os.environ[GOOGLE_APPLICATION_CREDENTIALS] = gke_creds
    #     pass


# def pytest_sessionfinish(session):
#     """Unconfigure the test session for kubetest.
#
#     See Also:
#         https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_sessionfinish
#     """
#     # Unset/reset the GOOGLE_APPLICATION_CREDENTIALS environment variable. For more,
#     # see: https://cloud.google.com/docs/authentication/production
#     gke_creds = config.getoption('google_application_credentials')
#     if gke_creds is not None:
#         # If there are old credentials stored, reset them. Otherwise, just unset
#         # the environment variable.
#         old = os.environ.get('OLD_' + GOOGLE_APPLICATION_CREDENTIALS)
#         if old:
#             os.environ[GOOGLE_APPLICATION_CREDENTIALS] = old
#             del os.environ['OLD_' + GOOGLE_APPLICATION_CREDENTIALS]
#         else:
#             del os.environ[GOOGLE_APPLICATION_CREDENTIALS]


def pytest_runtest_setup(item):
    """Run setup actions to prepare the test case.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_runtest_setup
    """
    # Previously (v0.3.[0-3]), this setup was gated to only run if the
    # --kube-config commandline flag was set. This was primarily done as an
    # optimization as to not create test metadata for tests that will not use
    # it. https://github.com/vapor-ware/kubetest/issues/130 points out that
    # the kubeconfig may be set without the --kube-config flag by overriding the
    # kubeconfig fixture. In such a case, this would have been skipped -- thus
    # there should NOT be any gating around test case metadata creation since
    # it is too early to tell whether we have all of the info we need.

    namespace_create = True
    namespace_name = None
    for mark in item.iter_markers(name="namespace"):
        namespace_create = mark.kwargs.get('create', True)
        namespace_name = mark.kwargs.get('name', None)

    # Register a new test case with the manager and setup the test case state.
    test_case = manager.new_test(
        node_id=item.nodeid,
        test_name=item.name,
        namespace_create=namespace_create,
        namespace_name=namespace_name,
    )

    # Note: These markers are not applied right now, meaning that the resource(s)
    #  which they reference are not added to the cluster yet. They are just
    #  registered with the test case so they can be applied to the cluster on
    #  test case setup.
    #
    #  At this point, the config is not loaded, so there is nothing that could be
    #  added to the cluster. It is safe to skip the teardown (which cleans up the
    #  test namespace) since nothing could be added to the namespace yet.
    try:
        # Register test case state based on markers on the test case.
        test_case.register_rolebindings(
            *markers.rolebindings_from_marker(item, test_case.ns)
        )
        test_case.register_clusterrolebindings(
            *markers.clusterrolebindings_from_marker(item, test_case.ns)
        )

        # Apply manifests for the test case, if any are specified.
        markers.apply_manifests_from_marker(item, test_case)
        markers.apply_manifest_from_marker(item, test_case)

    except Exception as e:
        test_case._pt_setup_failed = True
        raise e


def pytest_runtest_teardown(item):
    """Run teardown actions to clean up the test client.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_runtest_teardown
    """

    # See #154: Previously, this was only checking if the --kube-config command line
    # arg was passed to determine whether a test used kubetest for setup, and thus
    # whether it needed to be torn down. With the introduction of the kubeconfig
    # fixture, which allows users to override the kubeconfig in the test without having
    # to specify the --kube-config arg, this would prevent cleanup with the config was
    # specified via fixture.
    #
    # Unfortunately, due to the way the fixture is defined, we cannot simple check for
    # the presence/absence of the kubeconfig fixture in the test `item`, as it is always
    # present since there is a default kubeconfig fixture.
    #
    # As a bit of a hack, we can get the fixtures associated with the test item. When we
    # get the fixtures for `kubeconfig`, there will only be one fixture defined if no
    # custom fixture is specified - that is the default fixture. If there is a custom
    # override, there will be more than one fixture associated with the 'kubeconfig'
    # name. In such case, we should assume that a fixture was used to load the config
    # and allow test cleanup to proceed.
    _, _, fixtures = item.session._fixturemanager.getfixtureclosure(['kubeconfig'], item)

    if item.config.getoption('kube_config') or len(fixtures.get('kubeconfig', [])) > 1:
        manager.teardown(item.nodeid)


def pytest_runtest_makereport(item, call):
    """Create a test report for the test case. If the test case was found
    to fail, this will log out the container logs to provide more debugging
    context.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_runtest_makereport
    """

    # skip for tests without fixtures (eg: doctests)
    if not hasattr(item, 'fixturenames'):
        return

    if 'kube' in item.fixturenames and call.when == 'call':
        if call.excinfo is not None and call.excinfo.typename != 'Skipped':
            tail_lines = item.config.getoption('kube_error_log_lines')
            if tail_lines != 0:
                test_case = manager.get_test(item.nodeid)
                if test_case:
                    logs = test_case.yield_container_logs(
                        tail_lines=tail_lines
                    )
                    for container_log in logs:
                        # Add a report section to the test output
                        item.add_report_section(
                            when=call.when,
                            key='kubernetes container logs',
                            content=container_log
                        )


def pytest_keyboard_interrupt():
    """Clean up the cluster from kubetest artifacts if the tests
    are manually terminated via keyboard interrupt.

    See Also:
        https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_keyboard_interrupt
    """
    try:
        namespaces = kubernetes.client.CoreV1Api().list_namespace()
        for ns in namespaces.items:
            # if the namespace has a 'kubetest-' prefix, remove it.
            name = ns.metadata.name
            status = ns.status
            if (
                    name.startswith('kubetest-') and
                    status is not None and
                    status.phase.lower() == 'active'
            ):
                print('keyboard interrupt: cleaning up namespace "{}"'.format(name))
                kubernetes.client.CoreV1Api().delete_namespace(
                    body=kubernetes.client.V1DeleteOptions(),
                    name=name,
                )

        crbs = kubernetes.client.RbacAuthorizationV1Api().list_cluster_role_binding()
        for crb in crbs.items:
            # if the cluster role binding has a 'kubetest:' prefix, remove it.
            name = crb.metadata.name
            if name.startswith('kubetest:'):
                print(
                    'keyboard interrupt: cleaning up clusterrolebinding "{}"'.format(crb)
                )
                kubernetes.client.RbacAuthorizationV1Api().delete_cluster_role_binding(
                    body=kubernetes.client.V1DeleteOptions(),
                    name=name,
                )
    except Exception as e:
        log.error(
            'Failed to clean up kubetest artifacts from cluster on keyboard interrupt. '
            'You may need to manually remove items from your cluster. Check for '
            'namespaces with the "kubetest-" prefix and cluster role bindings with '
            'the "kubetest:" prefix. ({})'.format(e)
        )


# ********** pytest fixtures **********

class ClusterInfo:
    """Information about the cluster the kubetest is being run on.

    This info is gathered from the current context and the loaded
    configuration.

    :ivar cluster: The name of the cluster set for the current context.
    :ivar user: The name of the user set for the current context.
    :ivar context: The name of the current context.
    :ivar host: API server address.
    :ivar verify_ssl: SSL certificate verification when calling the API.
    """

    def __init__(self, current, config):
        self.cluster = current['context']['cluster']
        self.user = current['context']['user']
        self.context = current['name']
        self.host = config.host
        self.verify_ssl = config.verify_ssl


@pytest.fixture
def clusterinfo(request):
    """Get a ``ClusterInfo`` instance which provides basic information
    about the cluster the tests are being run on.
    """

    # Get an instance of a Configuration, which the kubeconfig
    # will be loaded into by default.
    config = kubernetes.client.Configuration()

    # Get the current context.
    _, current = kubernetes.config.list_kube_config_contexts(
        os.path.expandvars(os.path.expanduser(
            request.session.config.getoption('kube_config')
        ))
    )

    return ClusterInfo(
        current=current,
        config=config,
    )


@pytest.fixture
def kubeconfig(request):
    """Return the name of the configured kube config file loaded for the tests."""

    config_file = request.session.config.getoption('kube_config')
    return config_file


@pytest.fixture()
def kube(kubeconfig, request):
    """Return a client for managing a Kubernetes cluster for testing."""

    if not kubeconfig:
        log.error(
            'kube fixture used when no --kube-config is set; unable to install test '
            'resources onto a cluster.'
        )
        raise errors.SetupError('--kube-config not set')

    test_case = manager.get_test(request.node.nodeid)
    if test_case is None:
        log.error(
            'No kubetest client found for test using the "kube" fixture. ({})'.format(
                request.node.nodeid,
            )
        )
        raise errors.SetupError('error generating test client')

    # Setup the test case. This will create the namespace and any other
    # objects (e.g. role bindings) that the test case will need.
    kubernetes.config.load_kube_config(
        config_file=os.path.expandvars(os.path.expanduser(kubeconfig)),
        context=request.session.config.getoption('kube_context'),
    )
    test_case.setup()

    return test_case.client



"""Custom pytest markers for kubetest."""

import os

from kubernetes import client

from kubetest.manifest import load_file, load_path
from kubetest.objects import ApiObject, ClusterRoleBinding, RoleBinding

APPLYMANIFEST_INI = (
    'applymanifest(path): '
    'load a YAML manifest file from the specified path and create it on the cluster. '
    'This marker is similar to the "kubectl apply -f <path>" command. Loading a '
    'manifest via this marker will not prohibit you from loading other manifests '
    'manually. Use the "kube" fixture to get references to the created objects. The '
    'manifest loaded via this marker is registered with the internal test case '
    'metainfo and can be waited upon for creation via the "kube" fixture\'s '
    '"wait_until_created" method.'
)

APPLYMANIFESTS_INI = (
    'applymanifests(dir, files=None): '
    'load YAML manifests from the specified path and create them on the cluster. '
    'By default, all YAML files found in the specified path will be loaded and created. '
    'If a list is passed to the files parameter, only the files in the path matching '
    'a name in the files list will be loaded and created. This marker is similar to '
    'the "kubectl apply -f <dir>" command. Loading manifests via this marker will not '
    'prohibit you from loading other manifests manually. Use the "kube" fixture to get '
    'references to the created objects. Manifests loaded via this marker are registered '
    'with the internal test case metainfo and can be waited upon for creation via the '
    '"kube" fixture\'s "wait_until_created" method.'
)

CLUSTERROLEBINDING_INI = (
    'clusterrolebinding(name, subject_kind=None, subject_name=None): '
    'create and use a Kubernetes ClusterRoleBinding for the test case. The generated '
    'cluster role binding will be automatically created and removed for each marked '
    'test. The name of the role must be specified. Only existing ClusterRoles can be '
    'used. Optionally, the subject_kind (User, Group, ServiceAccount) and subject_name '
    'can be specified to set a target subject for the ClusterRoleBinding. If no '
    'subject is specified, it will default to all users in the namespace and all '
    'service accounts. If a subject is specified, both the subject kind and name must '
    'be present. The ClusterRoleBinding will always use the apiGroup '
    '"rbac.authorization.k8s.io" for both subjects and roleRefs. For more information, '
    'see: https://kubernetes.io/docs/reference/access-authn-authz/rbac/'
)

ROLEBINDING_INI = (
    'rolebinding(kind, name, subject_kind=None, subject_name=None): '
    'create and use a Kubernetes RoleBinding for the test case. The generated role '
    'binding will use the generated test-case namespace and will be automatically '
    'removed once the test completes. The role kind (Role, ClusterRole) must be '
    'specified along with the name of the role. Only existing Roles or ClusterRoles '
    'can be used. Optionally, the subject_kind (User, Group, ServiceAccount) and '
    'subject_name can be specified to set a target subject for the RoleBinding. If '
    'no subject is specified, it will default to all users in the namespace and all '
    'service accounts. If a subject is specified, both the subject kind and name '
    'must be present. The RoleBinding will always use the apiGroup '
    '"rbac.authorization.k8s.io" for both subjects and roleRefs. For more information, '
    'see: https://kubernetes.io/docs/reference/access-authn-authz/rbac/'
)


NAMESPACE_INI = (
    'namespace(create=True, name=None): '
    'finer-grained namespace configuration for the test case. By default, a new '
    'Kubernetes namespace is generated for each test case, using the test name and a '
    'timestamp to ensure uniqueness. This marker allows you to override this behavior '
    'to define your own namespace names, or to use existing namespaces. Set "create" '
    'to False to disable namespace  creation entirely. Set "name" to a string to set '
    'the name of the namespace to create/use.'
)


def register(config):
    """Register kubetest markers with pytest.

    Args:
        config: The pytest config that markers will be registered to.
    """
    config.addinivalue_line('markers', APPLYMANIFEST_INI)
    config.addinivalue_line('markers', APPLYMANIFESTS_INI)
    config.addinivalue_line('markers', CLUSTERROLEBINDING_INI)
    config.addinivalue_line('markers', ROLEBINDING_INI)
    config.addinivalue_line('markers', NAMESPACE_INI)


def apply_manifest_from_marker(item, client):
    """Load a manifest and create the API objects for teh specified file.

    This gets called for every `pytest.mark.applymanifest` marker on
    test cases.

    Once the manifest has been loaded, the API object(s) will be registered
    with the test cases' TestMeta. This allows easier control via the
    "kube" fixture, such as waiting for all objects to be created.

    Args:
        item: The pytest test item.
        client (manager.TestMeta): The metainfo object for the marked
            test case.
    """
    for mark in item.iter_markers(name='applymanifest'):
        path = mark.args[0]

        # Normalize the path to be absolute.
        if not os.path.isabs(path):
            path = os.path.abspath(path)

        # Load the manifest
        objs = load_file(path)

        # For each of the loaded Kubernetes resources, wrap it in the
        # equivalent kubetest wrapper. If the object does not yet have a
        # wrapper, error out. We cannot reliably create the resource
        # without our ApiObject wrapper semantics.
        wrapped = []
        found = False
        for obj in objs:
            for klass in ApiObject.__subclasses__():
                if obj.kind == klass.__name__:
                    wrapped.append(klass(obj))
                    found = True
                    break
            if not found:
                raise ValueError(
                    'Unable to match loaded object to an internal wrapper '
                    'class: {}'.format(obj)
                )

        client.register_objects(wrapped)


def apply_manifests_from_marker(item, client):
    """Load manifests and create the API objects for the specified files.

    This gets called for every `pytest.mark.applymanifests` marker on
    test cases.

    Once a manifest has been loaded, the API objects will be registered
    with the test cases' TestMeta. This allows some easier control via
    the "kube" fixture, such as waiting for all objects to be created.

    Args:
        item: The pytest test item.
        client (manager.TestMeta): The metainfo object for the marked
            test case.
    """
    for mark in item.iter_markers(name='applymanifests'):
        dir_path = mark.args[0]
        files = mark.kwargs.get('files')

        # We expect the path specified to either be absolute or relative
        # from the test file. If the path is relative, add the directory
        # that the test file resides in as a prefix to the dir_path.
        if not os.path.isabs(dir_path):
            dir_path = os.path.abspath(
                os.path.join(os.path.dirname(item.fspath), dir_path)
            )

        # If there are any files specified, we will only load those files.
        # Otherwise, we'll load everything in the directory.
        if files is None:
            objs = load_path(dir_path)
        else:
            objs = []
            for f in files:
                objs.extend(load_file(os.path.join(dir_path, f)))

        # For each of the loaded Kubernetes resources, we'll want to wrap it
        # in the equivalent kubetest wrapper. If the resource does not have
        # an equivalent kubetest wrapper, error out. We cannot reliably create
        # the resource without our ApiObject wrapper semantics.
        wrapped = []
        for obj in objs:
            found = False
            for klass in ApiObject.__subclasses__():
                if obj.kind == klass.__name__:
                    wrapped.append(klass(obj))
                    found = True
                    break
            if not found:
                raise ValueError(
                    'Unable to match loaded object to an internal wrapper '
                    'class: {}'.format(obj)
                )

        client.register_objects(wrapped)


def rolebindings_from_marker(item, namespace):
    """Create RoleBindings for the test case if the test is marked
    with the `pytest.mark.rolebinding` marker.

    Args:
        item (pytest.Item): The pytest test item.
        namespace (str): The namespace of the test case.

    Returns:
        list[objects.RoleBinding]: The RoleBindings that were
            generated from the test case markers.
    """
    rolebindings = []
    for mark in item.iter_markers(name='rolebinding'):
        kind = mark.args[0]
        name = mark.args[1]
        subj_kind = mark.kwargs.get('subject_kind')
        subj_name = mark.kwargs.get('subject_name')

        subj = get_custom_rbac_subject(namespace, subj_kind, subj_name)
        if not subj:
            subj = get_default_rbac_subjects(namespace)

        rolebindings.append(RoleBinding(client.V1RoleBinding(
            metadata=client.V1ObjectMeta(
                name='kubetest:{}'.format(item.name),
                namespace=namespace,
            ),
            role_ref=client.V1RoleRef(
                api_group='rbac.authorization.k8s.io',
                kind=kind,
                name=name,
            ),
            subjects=subj,
        )))

    return rolebindings


def clusterrolebindings_from_marker(item, namespace):
    """Create ClusterRoleBindings for the test case if the test case
    is marked with the `pytest.mark.clusterrolebinding` marker.

    Args:
        item (pytest.Item): The pytest test item.
        namespace (str): The namespace of the test case.

    Return:
        list[objects.ClusterRoleBinding]: The ClusterRoleBindings that
            were generated from the test case markers.
    """
    clusterrolebindings = []
    for mark in item.iter_markers(name='clusterrolebinding'):
        name = mark.args[0]
        subj_kind = mark.kwargs.get('subject_kind')
        subj_name = mark.kwargs.get('subject_name')

        subj = get_custom_rbac_subject(namespace, subj_kind, subj_name)
        if not subj:
            subj = get_default_rbac_subjects(namespace)

        clusterrolebindings.append(ClusterRoleBinding(client.V1ClusterRoleBinding(
            metadata=client.V1ObjectMeta(
                name='kubetest:{}'.format(item.name),
            ),
            role_ref=client.V1RoleRef(
                api_group='rbac.authorization.k8s.io',
                kind='ClusterRole',
                name=name,
            ),
            subjects=subj,
        )))

    return clusterrolebindings


def get_custom_rbac_subject(namespace, kind, name):
    """Create a custom RBAC subject for the given namespace.

    Both `kind` and `name` must be specified. If one is set and
    the other is not (None), an error will be raised.

    Args:
        namespace (str): The namespace of the Subject.
        kind (str): The subject kind. This should be one of:
            'User', 'Group', or 'ServiceAccount'.
        name (str): The name of the Subject.

    Returns:
        list[client.V1Subject]: The custom RBAC subject.

    Raises:
        ValueError: One of `kind` and `name` are None.
    """
    # check that both `kind` and `name` are set.
    if (kind and not name) or (not kind and name):
        raise ValueError(
            'One of subject_kind and subject_name were specified, but both must '
            'be specified when defining a custom Subject.'
        )

    # otherwise, if a custom subject is specified, create it
    if name is not None and kind is not None:
        return [
            client.V1Subject(
                api_group='rbac.authorization.k8s.io',
                namespace=namespace,
                kind=kind,
                name=name,
            )
        ]
    else:
        return []


def get_default_rbac_subjects(namespace):
    """Get the default RBAC Subjects.

    The default subjects will allow:
      - all authenticated users
      - all unauthenticated users
      - all service accounts

    Args:
        namespace (str): The namespace of the Subjects.

    Returns:
        list[client.V1Subject]: The default RBAC subjects.
    """
    return [
        # all authenticated users
        client.V1Subject(
            api_group='rbac.authorization.k8s.io',
            namespace=namespace,
            name='system:authenticated',
            kind='Group',
        ),
        # all unauthenticated users
        client.V1Subject(
            api_group='rbac.authorization.k8s.io',
            namespace=namespace,
            name='system:unauthenticated',
            kind='Group',
        ),
        # all service accounts
        client.V1Subject(
            api_group='rbac.authorization.k8s.io',
            namespace=namespace,
            name='system:serviceaccounts',
            kind='Group',
        ),
    ]



"""kubetest -- a Kubernetes integration test framework in Python."""

__title__ = 'kubetest'
__version__ = '0.5.1'
__description__ = 'A Kubernetes integration test framework in Python.'
__author__ = 'Vapor IO'
__author_email__ = 'vapor@vapor.io'
__url__ = 'https://github.com/vapor-ware/kubetest'
__license__ = 'GNU General Public License v3.0'



"""Kubetest wrapper for the Kubernetes ``RoleBinding`` API Object."""

import logging

from kubernetes import client

from .api_object import ApiObject

log = logging.getLogger('kubetest')


class RoleBinding(ApiObject):
    """Kubetest wrapper around a Kubernetes `RoleBinding`_ API Object.

    The actual ``kubernetes.client.V1RoleBinding`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `RoleBinding`_.

    .. _RoleBinding:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#rolebinding-v1-rbac-authorization-k8s-io
    """

    obj_type = client.V1RoleBinding

    api_clients = {
        'preferred': client.RbacAuthorizationV1Api,
        'rbac.authorization.k8s.io/v1': client.RbacAuthorizationV1Api,
        'rbac.authorization.k8s.io/v1alpha1': client.RbacAuthorizationV1alpha1Api,
        'rbac.authorization.k8s.io/v1beta1': client.RbacAuthorizationV1beta1Api,
    }

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def create(self, namespace=None):
        """Create the RoleBinding under the given namespace.

        Args:
            namespace (str): The namespace to create the RoleBinding under.
                If the RoleBinding was loaded via the kubetest client, the
                namespace will already be set, so it is not needed here.
                Otherwise, the namespace will need to be provided.
        """
        if namespace is None:
            namespace = self.namespace

        log.info('creating rolebinding "%s" in namespace "%s"', self.name, self.namespace)  # noqa
        log.debug('rolebinding: %s', self.obj)

        self.obj = self.api_client.create_namespaced_role_binding(
            namespace=namespace,
            body=self.obj,
        )

    def delete(self, options):
        """Delete the RoleBinding.

        This method expects the RoleBinding to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will need
        to be set manually.

        Args:
             options (client.V1DeleteOptions): Options for RoleBinding deletion.

        Returns:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting rolebinding "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('rolebinding: %s', self.obj)

        return self.api_client.delete_namespaced_role_binding(
            namespace=self.namespace,
            name=self.name,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes RoleBinding resource."""
        self.obj = self.api_client.read_namespaced_role_binding(
            namespace=self.namespace,
            name=self.name,
        )

    def is_ready(self):
        """Check if the RoleBinding is in the ready state.

        RoleBindings do not have a "status" field to check, so we
        will measure their readiness status by whether or not they exist
        on the cluster.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        try:
            self.refresh()
        except:  # noqa
            return False
        else:
            return True



"""Response wrapper object for proxied requests to Kubernetes resources."""

import ast
import json
import logging

import urllib3

log = logging.getLogger('kubetest')


class Response:
    """Response is a wrapper around the Kubernetes API's response data
    when a request is proxied to a Kubernetes resource, like a Pod or
    Service.

    A proxied request will return:
    - The response data
    - The response headers
    - The response HTTP code

    All of these are wrapped by this class. Additional helpers are provided,
    such as casting the response data to JSON.

    Args:
        data: The response data from the proxied request.
        status: The response status code.
        headers: The response headers.
    """

    def __init__(self, data, status, headers):
        self.data = data
        self.status = status
        self.headers = headers

    def json(self):
        """Convert the response data to JSON.

        If the response data is not valid JSON, an error will be raised.

        Returns:
            dict: The JSON data.
        """
        # This should generally be the case, since the primary source of Response
        # instances are the HTTP methods on Pods (e.g. http_proxy_get), where
        # `_preload_content=False`. Should that be set to True, it will not return
        # the HTTPResponse but will instead return the response data formatted as
        # the type specified in the `response_type` param. By default, kubetest sets
        # the _preload_content field to True, so this should generally not be hit.
        if isinstance(self.data, urllib3.HTTPResponse):
            return json.loads(self.data.data)

        # The response data comes back as a string. This could be a JSON string,
        # or something else (text body, error string, etc). Since we've preloaded
        # the content as a string (see comment above), we can not simply load the
        # string as JSON as the preloading essentially serializes the Python dict
        # out to a string, so various values will not parse into JSON (None vs null,
        # " vs ', etc). To remedy this, we attempt to load the response as an ast
        # literal - if it fails for any reason, fall back to trying to load JSON.
        # At the very least the JSON loading will fail with a familiar error which
        # one would expect from a .json() function.
        try:
            data = ast.literal_eval(self.data)
        except Exception as e:
            log.debug('failed literal eval of data {} ({})'.format(self.data, e))
            data = json.loads(self.data)

        return data



"""Utility functions for loading Kubernetes manifest files and constructing
the corresponding Kubernetes API models.
"""

import builtins
import os
import re

import kubernetes
import yaml
from kubernetes.client import models


def load_file(path):
    """Load an individual Kubernetes manifest YAML file.

    This file may contain multiple YAML documents. It will attempt to auto-detect
    the type of each object to load.

    Args:
        path (str): The fully qualified path to the file.

    Returns:
        A list of the Kubernetes API objects for this manifest file.
    """
    with open(path, 'r') as f:
        manifests = yaml.load_all(f, Loader=yaml.SafeLoader)

        objs = []
        for manifest in manifests:
            obj_type = get_type(manifest)
            if obj_type is None:
                raise ValueError(
                    'Unable to determine object type for manifest: {}'.format(manifest)
                )
            objs.append(new_object(obj_type, manifest))

    return objs


def load_path(path):
    """Load all of the Kubernetes YAML manifest files found in the
    specified directory path.

    Args:
        path (str): The path to the directory of manifest files.

    Returns:
        list: A list of all the Kubernetes objects loaded from
            manifest file.

    Raises:
        ValueError: The provided path is not a directory.
    """
    if not os.path.isdir(path):
        raise ValueError('{} is not a directory'.format(path))

    objs = []
    for f in os.listdir(path):
        if os.path.splitext(f)[1].lower() in ['.yaml', '.yml']:
            objs = objs + load_file(os.path.join(path, f))
    return objs


def get_type(manifest):
    """Get the Kubernetes object type from the manifest kind and
    version.

    There is no easy way for determining the internal model that a
    manifest should use. What this tries to do is use the version
    info and the kind info to create a potential internal object
    name for the pair and look that up in the kubernetes package
    locals.

    Args:
        manifest (dict): The manifest file, loaded into a dictionary.

    Returns:
        object: The Kubernetes API object for the manifest.
        None: No Kubernetes API object type could be determined.

    Raises:
        ValueError: The manifest dictionary does not have a
            `version` or `kind` specified.
    """
    version = manifest.get('apiVersion')
    if version is None:
        raise ValueError('manifest has no "version" field specified')

    kind = manifest.get('kind')
    if kind is None:
        raise ValueError('manifest has no "kind" field specified')

    # create a map of the kubernetes client model locals where the key is the
    # lower cased name (so we don't have to mess with getting the capitalization
    # of components correct) and the value is the correctly cased name.
    lookup = {k.lower(): k for k in models.__dict__.keys()}

    # if the version has a '/' (e.g. apps/v1, extensions/v1beta1), remove it.

    # there are generally two possibilities - we include the version minus
    # the slash (e.g. extensions/v1beta1 -> extensionsv1beta1), or we do not
    # use the prefix (e.g. apps/v1 -> v1). By default we will always try with
    # the prefix first and only try the secondary check if the first check
    # yields nothing.
    possibilities = [
        # add the default case
        version.replace('/', '').replace('.', '') + kind
    ]

    # if the prefix exists, add the non-prefixed version as a secondary check
    if version.count('/') == 1:
        possibilities.append(version.split('/')[1] + kind)

    for to_check in possibilities:
        type_name = lookup.get(to_check.lower())
        if type_name is None:
            continue
        return models.__dict__.get(type_name)
    return None


def load_type(obj_type, path):
    """Load a Kubernetes YAML manifest file for the specified type.

    While Kubernetes manifests can contain multiple object definitions
    in a single file (delimited with the YAML separator '---'), this
    does not currently support those files. This function expects a
    single object definition in the specified manifest file.

    Args:
        path (str): The path the manifest YAML to load.
        obj_type: The Kubernetes API object type that the YAML
            contents should be loaded into.

    Returns:
        A Kubernetes API object populated with the YAML contents.

    Raises:
        FileNotFoundError: The specified file was not found.
    """
    with open(path, 'r') as f:
        manifest = yaml.full_load(f)

    return new_object(obj_type, manifest)


def new_object(root_type, config):
    """Create a new Kubernetes API object and recursively populate it with
    the provided manifest configuration.

    The recursive population utilizes the swagger_types and attribute_map
    members of the Kubernetes API object class to determine which config
    fields correspond to which input parameter, and to cast them to their
    expected type.

    This is all based on the premise that the Python Kubernetes client will
    continue to be based off of an auto-generated Swagger spec and that these
    fields will be available for all API objects.

    Args:
        root_type: The Kubernetes API object type that will be populated
            with the manifest configuration. This is expected to be known
            ahead of time by the caller.
        config: The manifest configuration for the API object.

    Returns:
        A Kubernetes API object recursively populated with the YAML contents.
    """

    # The arguments that will be passed to the root_type to create a new
    # recursively populated instance of that type.
    constructor_args = {}

    # The attribute map maps the argument name (e.g. api_version) to the name
    # of the corresponding configuration field (e.g. apiVersion). Iterate over
    # each of these to pick up all the possible configuration options from the
    # provided manifest.
    for k, v in root_type.attribute_map.items():
        cfg_value = config.get(v)
        if cfg_value is not None:

            # The config value matches an expected key in the attribute dict.
            # Now, we want to cast that config to the appropriate type based
            # on the contents of the 'swagger_types' dict.
            t = root_type.swagger_types[k]

            # There are two classes of types we will want to check against:
            # 'base types' (like: str, int, etc) and 'collection types'
            # (like: list, dict). Collection types can contain base types,
            # so we will want to apply the same base type checks to each
            # element within a collection type. First we will check for the
            # collection types. If it is neither, we assume that the type is
            # a base type.

            # Check if the type is a list of some other type.
            # This should match to something like: 'list[str]', where the
            # element type (in this case 'str') will be isolated as a group.
            list_match = re.match(r'^list\[(.*)\]$', t)
            if list_match is not None:
                element_type = list_match.group(1)
                list_value = [cast_value(i, element_type) for i in cfg_value]
                constructor_args[k] = list_value
                continue

            # Check if the type is a dict composed of other types.
            # This should match to something lint: 'dict(str, str)', where
            # the element types (in this case, both 'str') will be isolated
            # as separate groups.
            dict_match = re.match(r'^dict\((.*), (.*)\)$', t)
            if dict_match is not None:
                key_type = dict_match.group(1)
                val_type = dict_match.group(2)
                dict_value = {
                    cast_value(k, key_type): cast_value(v, val_type)
                    for k, v in cfg_value.items()
                }
                constructor_args[k] = dict_value
                continue

            # If it is not a collection type, it must be a base type.
            constructor_args[k] = cast_value(cfg_value, t)

    return root_type(**constructor_args)


def cast_value(value, t):
    """Cast the given value to the specified type.

    There are two general cases for possible casts:
      - A cast to a builtin type (int, str, etc.)
      - A cast to a Kubernetes object (V1ObjectMeta, etc)

    In either case, check to see if the specified type exists in the
    correct type pool. If so, cast to that type, otherwise fail.

    Args:
        value: The value to cast.
        t (str): The type to cast the value to. This can be a builtin
            type or a Kubernetes API object type.

    Returns:
        The value, casted to the appropriate type.

    Raises:
        ValueError: Unable to cast the value to the specified type.
        TypeError: Unable to cast the given value to the specified type.
        AttributeError: The value is an invalid Kubernetes type.
    """

    # The config value should be cast to a built-in type
    builtin_type = builtins.__dict__.get(t)
    if builtin_type == object:
        return value
    if builtin_type is not None:
        return builtin_type(value)

    # The config value should be cast to a Kubernetes type
    k_type = kubernetes.client.__dict__.get(t)
    if k_type is not None:
        return new_object(k_type, value)

    raise ValueError('Unable to determine cast type behavior: {}'.format(t))



"""Kubetest wrapper for the Kubernetes `Node` API Object."""

import logging

from kubernetes import client

log = logging.getLogger('kubetest')


class Node:
    """Kubetest wrapper around a Kubernetes `Node`_ API Object.

    The actual ``kubernetes.client.V1Node`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `Node`_.

    This wrapper does **NOT** subclass the ``objects.ApiObject`` like
    other object wrappers because it is not intended to be created or
    managed from manifest file. It is merely meant to wrap the
    Node spec to make Node-based interactions easier

    .. _Node:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#node-v1-core
    """

    def __init__(self, api_object):
        self.obj = api_object
        self.name = api_object.metadata.name

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def refresh(self):
        """Refresh the underlying Kubernetes Node resource."""
        nodes = client.CoreV1Api().list_node()
        for node in nodes.items:
            if node.metadata.name == self.name:
                self.obj = node
                return
        log.warning('unable to refresh node: no node found with name: %s', self.name)

    def status(self):
        """Get the status of the Node.

        Returns:
            client.V1NodeStatus: The status of the Node.
        """
        log.info('checking status of node "%s"', self.name)
        self.refresh()
        return self.obj.status

    def is_ready(self):
        """Check whether the Node is in the ready state.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        status = self.status()

        # if there is no status, the node is definitely not ready
        if status is None:
            return False

        # if there are no conditions set, the node is not ready
        if status.conditions is None:
            return False

        for cond in status.conditions:
            # we only care about the 'ready' condition
            if cond.type.lower() != 'ready':
                continue

            # check that the readiness condition is true
            return cond.status.lower() == 'true'

        # Catchall
        return False



"""Kubetest wrapper for the Kubernetes `Pod` API Object."""

import logging

from kubernetes import client
from kubernetes.client.rest import ApiException

from kubetest import condition, response, utils

from .api_object import ApiObject
from .container import Container

log = logging.getLogger('kubetest')


class Pod(ApiObject):
    """Kubetest wrapper around a Kubernetes `Pod`_ API Object.

    The actual ``kubernetes.client.V1Pod`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `Pod`_.

    .. _Pod:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#pod-v1-core
    """

    obj_type = client.V1Pod

    api_clients = {
        'preferred': client.CoreV1Api,
        'v1': client.CoreV1Api,
    }

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def create(self, namespace=None):
        """Create the Pod under the given namespace.

        Args:
            namespace (str): The namespace to create the Pod under.
                If the Pod was loaded via the kubetest client, the
                namespace will already be set, so it is not needed
                here. Otherwise, the namespace will need to be provided.
        """
        if namespace is None:
            namespace = self.namespace

        log.info('creating pod "%s" in namespace "%s"', self.name, self.namespace)
        log.debug('pod: %s', self.obj)

        self.obj = self.api_client.create_namespaced_pod(
            namespace=namespace,
            body=self.obj,
        )

    def delete(self, options):
        """Delete the Pod.

        This method expects the Pod to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will
        need to be set manually.

        Args:
            options (client.V1DeleteOptions): Options for Pod deletion.

        Return:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting pod "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('pod: %s', self.obj)

        return self.api_client.delete_namespaced_pod(
            name=self.name,
            namespace=self.namespace,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes Pod resource."""
        self.obj = self.api_client.read_namespaced_pod_status(
            name=self.name,
            namespace=self.namespace,
        )

    def is_ready(self):
        """Check if the Pod is in the ready state.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        self.refresh()

        # if there is no status, the pod is definitely not ready
        status = self.obj.status
        if status is None:
            return False

        # check the pod phase to make sure it is running. a pod in
        # the 'failed' or 'success' state will no longer be running,
        # so we only care if the pod is in the 'running' state.
        phase = status.phase
        if phase.lower() != 'running':
            return False

        for cond in status.conditions:
            # we only care about the condition type 'ready'
            if cond.type.lower() != 'ready':
                continue

            # check that the readiness condition is True
            return cond.status.lower() == 'true'

        # Catchall
        return False

    def status(self):
        """Get the status of the Pod.

        Returns:
            client.V1PodStatus: The status of the Pod.
        """
        # first, refresh the pod state to ensure latest status
        self.refresh()

        # return the status of the pod
        return self.obj.status

    def get_containers(self):
        """Get the Pod's containers.

        Returns:
            list[Container]: A list of containers that belong to the Pod.
        """
        log.info('getting containers for pod "%s"', self.name)
        self.refresh()

        return [Container(c, self) for c in self.obj.spec.containers]

    def get_container(self, name):
        """Get a container in the Pod by name.

        Args:
            name (str): The name of the Container.

        Returns:
            Container: The Pod's Container with the matching name. If
            no container with the given name is found, ``None`` is returned.
        """
        for c in self.get_containers():
            if c.obj.name == name:
                return c
        return None

    def get_restart_count(self):
        """Get the total number of Container restarts for the Pod.

        Returns:
            int: The total number of Container restarts.
        """
        container_statuses = self.status().container_statuses
        if container_statuses is None:
            return 0

        total = 0
        for container_status in container_statuses:
            total += container_status.restart_count

        return total

    def http_proxy_get(self, path, query_params=None):
        """Issue a GET request to a proxy for the Pod.

        Notes:
            This function does not use the kubernetes
            ``connect_get_namespaced_pod_proxy_with_path`` function because there
            appears to be lack of support for custom query parameters (as of
            the ``kubernetes==9.0.0`` package version). To bypass this, parts of
            the core functionality from the aforementioned function are used here with
            the modification of allowing user-defined query parameters to be
            passed along.

        Args:
            path (str): The URI path for the request.
            query_params (dict[str, str]): Any query parameters for
                the request. (default: None)

        Returns:
            response.Response: The response data.
        """
        c = client.CoreV1Api()

        if query_params is None:
            query_params = {}

        path_params = {
            'name': self.name,
            'namespace': self.namespace
        }
        header_params = {
            'Accept': c.api_client.select_header_accept(['*/*']),
            'Content-Type': c.api_client.select_header_content_type(['*/*'])
        }
        auth_settings = ['BearerToken']

        try:
            resp = response.Response(*c.api_client.call_api(
                '/api/v1/namespaces/{namespace}/pods/{name}/proxy/' + path, 'GET',
                path_params=path_params,
                query_params=query_params,
                header_params=header_params,
                body=None,
                post_params=[],
                files={},
                response_type='str',
                auth_settings=auth_settings,
                _return_http_data_only=False,  # we want all info, not just data
                _preload_content=True,
                _request_timeout=None,
                collection_formats={}
            ))
        except ApiException as e:
            # if the ApiException does not have a body or headers, that
            # means the raised exception did not get a response (even if
            # it were 404, 500, etc), so we want to continue to raise in
            # that case. if there is a body and headers, we will not raise
            # and just take the data out that we need from the exception.
            if e.body is None and e.headers is None:
                raise

            resp = response.Response(
                data=e.body,
                status=e.status,
                headers=e.headers,
            )

        return resp

    def http_proxy_post(self, path, query_params=None, data=None):
        """Issue a POST request to a proxy for the Pod.

        Notes:
            This function does not use the kubernetes
            ``connect_post_namespaced_pod_proxy_with_path`` function because there
            appears to be lack of support for custom query parameters (as of
            the ``kubernetes==9.0.0`` package version). To bypass this, parts of
            the core functionality from the aforementioned function are used here with
            the modification of allowing user-defined query parameters to be
            passed along.

        Args:
            path (str): The URI path for the request.
            query_params (dict[str, str]): Any query parameters for
                the request. (default: None)
            data: The data to POST.

        Returns:
            response.Response: The response data.
        """
        c = client.CoreV1Api()

        if query_params is None:
            query_params = {}

        path_params = {
            'name': self.name,
            'namespace': self.namespace
        }
        header_params = {
            'Accept': c.api_client.select_header_accept(['*/*']),
            'Content-Type': c.api_client.select_header_content_type(['*/*'])
        }
        auth_settings = ['BearerToken']

        try:
            resp = response.Response(*c.api_client.call_api(
                '/api/v1/namespaces/{namespace}/pods/{name}/proxy/' + path, 'POST',
                path_params=path_params,
                query_params=query_params,
                header_params=header_params,
                body=data,
                post_params=[],
                files={},
                response_type='str',
                auth_settings=auth_settings,
                _return_http_data_only=False,  # we want all info, not just data
                _preload_content=True,
                _request_timeout=None,
                collection_formats={}
            ))
        except ApiException as e:
            # if the ApiException does not have a body or headers, that
            # means the raised exception did not get a response (even if
            # it were 404, 500, etc), so we want to continue to raise in
            # that case. if there is a body and headers, we will not raise
            # and just take the data out that we need from the exception.
            if e.body is None and e.headers is None:
                raise

            resp = response.Response(
                data=e.body,
                status=e.status,
                headers=e.headers,
            )

        return resp

    def containers_started(self):
        """Check if the Pod's Containers have all started.

        Returns:
            bool: True if all Containers have started; False otherwise.
        """
        # start the flag as true - we will check the state and set
        # this to False if any container is not yet running.
        containers_started = True

        status = self.status()
        if status.container_statuses is not None:
            for container_status in status.container_statuses:
                if container_status.state is not None:
                    if container_status.state.running is not None:
                        if container_status.state.running.started_at is not None:
                            # The container is started, so move on to check the
                            # next container
                            continue
                # If we get here, then the container has not started.
                containers_started = containers_started and False
                break

        return containers_started

    def wait_until_containers_start(self, timeout=None):
        """Wait until all containers in the Pod have started.

        This will wait for the images to be pulled and for the containers
        to be created and started. This will unblock once all Pod containers
        have been started.

        This is different than waiting until ready, since a container may
        not be ready immediately after it has been started.

        Args:
            timeout (int): The maximum time to wait, in seconds, for the
                Pod's containers to be started. If unspecified, this will
                wait indefinitely. If specified and the timeout is met or
                exceeded, a TimeoutError will be raised.

        Raises:
            TimeoutError: The specified timeout was exceeded.
        """
        wait_condition = condition.Condition(
            'all pod containers started',
            self.containers_started,
        )

        utils.wait_for_condition(
            condition=wait_condition,
            timeout=timeout,
            interval=1,
        )



"""Kubetest wrapper for the Kubernetes ``ConfigMap`` API Object."""

import logging

from kubernetes import client

from .api_object import ApiObject

log = logging.getLogger('kubetest')


class ConfigMap(ApiObject):
    """Kubetest wrapper around a Kubernetes `ConfigMap`_ API Object.

    The actual ``kubernetes.client.V1ConfigMap`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `ConfigMap`_.

    .. _ConfigMap:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#configmap-v1-core
    """

    obj_type = client.V1ConfigMap

    api_clients = {
        'preferred': client.CoreV1Api,
        'v1': client.CoreV1Api,
    }

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def create(self, namespace=None):
        """Create the ConfigMap under the given namespace.

        Args:
            namespace (str): The namespace to create the ConfigMap under.
                If the ConfigMap was loaded via the kubetest client, the
                namespace will already be set, so it is not needed here.
                Otherwise, the namespace will need to be provided.
        """
        if namespace is None:
            namespace = self.namespace

        log.info('creating configmap "%s" in namespace "%s"', self.name, self.namespace)
        log.debug('configmap: %s', self.obj)

        self.obj = self.api_client.create_namespaced_config_map(
            namespace=namespace,
            body=self.obj,
        )

    def delete(self, options):
        """Delete the ConfigMap.

        This method expects the ConfigMap to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will need
        to be set manually.

        Args:
             options (client.V1DeleteOptions): Options for ConfigMap deletion.

        Returns:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting configmap "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('configmap: %s', self.obj)

        return self.api_client.delete_namespaced_config_map(
            name=self.name,
            namespace=self.namespace,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes ConfigMap resource."""
        self.obj = self.api_client.read_namespaced_config_map(
            name=self.name,
            namespace=self.namespace,
        )

    def is_ready(self):
        """Check if the ConfigMap is in the ready state.

        ConfigMaps do not have a "status" field to check, so we will
        measure their readiness status by whether or not they exist
        on the cluster.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        try:
            self.refresh()
        except:  # noqa
            return False
        else:
            return True



"""Utility functions for kubetest."""

import logging
import time

from kubernetes.client.rest import ApiException

log = logging.getLogger('kubetest')


def new_namespace(test_name):
    """Create a new namespace for the given test name.

    Kubernetes namespace names follow a DNS-1123 label that consists
    of lower case alphanumeric characters or '-' and must start with
    an alphanumeric.

    The test name and current timestamp are formatted to comply to
    this spec and appended to the 'kubetest' prefix.

    Args:
        test_name (str): The name of the test case for the namespace.

    Returns:
        str: The namespace name.
    """
    prefix = 'kubetest'
    timestamp = str(int(time.time()))
    test_name = test_name.replace('_', '-').lower()
    test_name = test_name.replace('[', '-')
    test_name = test_name.replace(']', '-')

    # The length of a resource name in Kubernetes may not exceed 63
    # characters. Check the length of all components (+2 for the dashes
    # joining the components). If the total length exceeds 63, truncate
    # the test name.
    name_len = len(prefix) + len(timestamp) + len(test_name) + 2

    if name_len > 63:
        test_name = test_name[:-(name_len-63)]

    return '-'.join((prefix, test_name, timestamp))


def selector_string(selectors):
    """Create a selector string from the given dictionary of selectors.

    Args:
        selectors (dict): The selectors to stringify.

    Returns:
        str: The selector string for the given dictionary.
    """
    return ','.join(['{}={}'.format(k, v) for k, v in selectors.items()])


def selector_kwargs(fields=None, labels=None):
    """Create a dictionary of kwargs for Kubernetes object selectors.

    Args:
        fields (dict[str, str]): A dictionary of fields used to restrict
            the returned collection of Objects to only those which match
            these field selectors. By default, no restricting is done.
        labels (dict[str, str]): A dictionary of labels used to restrict
            the returned collection of Objects to only those which match
            these label selectors. By default, no restricting is done.

    Returns:
        dict[str, str]: A dictionary that can be used as kwargs for
            many Kubernetes API calls for label and field selectors.
    """
    kwargs = {}
    if fields is not None:
        kwargs['field_selector'] = selector_string(fields)
    if labels is not None:
        kwargs['label_selector'] = selector_string(labels)

    return kwargs


def wait_for_condition(condition, timeout=None, interval=1, fail_on_api_error=True):
    """Wait for a condition to be met.

    Args:
        condition (condition.Condition): The Condition to wait for.
        timeout (int): The maximum time to wait, in seconds, for the
            condition to be met. If unspecified, this function will
            wait indefinitely. If specified and the timeout is met
            or exceeded, a TimeoutError will be raised.
        interval (int|float): The time, in seconds, to wait before
            re-checking the condition.
        fail_on_api_error (bool): Fail the condition checks if a Kubernetes
            API error is incurred. An API error can be raised for a number
            of reasons, including a Pod being restarted and temporarily
            unavailable. Disabling this will cause those errors to be
            ignored, allowing the check to continue until timeout or
            resolution. (default: True).

    Raises:
        TimeoutError: The specified timeout was exceeded.
    """
    log.info('waiting for condition: %s', condition)

    # define the maximum time to wait. once this is met, we should
    # stop waiting.
    max_time = None
    if timeout is not None:
        max_time = time.time() + timeout

    # start the wait block
    start = time.time()
    while True:
        if max_time and time.time() >= max_time:
            raise TimeoutError(
                'timed out ({}s) while waiting for condition {}'
                .format(timeout, condition)
            )

        # check if the condition is met and break out if it is
        try:
            if condition.check():
                break
        except ApiException as e:
            log.warning('got api exception while waiting: {}'.format(e))
            if fail_on_api_error:
                raise

        # if the condition is not met, sleep for the interval
        # to re-check later
        time.sleep(interval)

    end = time.time()
    log.info('wait completed (total=%fs) %s', end - start, condition)



"""Kubetest wrappers around Kubernetes API Objects."""

# flake8: noqa

from .api_object import ApiObject
from .clusterrolebinding import ClusterRoleBinding
from .configmap import ConfigMap
from .container import Container
from .daemonset import DaemonSet
from .deployment import Deployment
from .endpoints import Endpoints
from .event import Event
from .namespace import Namespace
from .node import Node
from .pod import Pod
from .rolebinding import RoleBinding
from .secret import Secret
from .service import Service
from .statefulset import StatefulSet



"""Kubetest wrapper for the Kubernetes ``Event`` API Object."""

import logging

from kubernetes import client

log = logging.getLogger('kubetest')


class Event:
    """Kubetest wrapper around a Kubernetes `Event`_ API Object.

    The actual ``kubernetes.client.V1Event`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper does **NOT** subclass the ``objects.ApiObject`` like
    other object wrappers because it is not intended to be created or
    managed from manifest file. It is merely meant to wrap the
    Event object to make Event-based interactions easier

    .. _Event:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#event-v1-core
    """

    obj_type = client.V1Event

    api_clients = {
        'preferred': client.CoreV1Api,
        'v1': client.CoreV1Api,
    }

    def __init__(self, api_object):
        self.obj = api_object
        self.name = api_object.metadata.name

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()



"""Kubetest wrapper for the Kubernetes ``ClusterRoleBinding`` API Object."""

import logging

from kubernetes import client

from .api_object import ApiObject

log = logging.getLogger('kubetest')


class ClusterRoleBinding(ApiObject):
    """Kubetest wrapper around a Kubernetes `ClusterRoleBinding`_ API Object.

    The actual ``kubernetes.client.V1ClusterRoleBinding`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `ClusterRoleBinding`_.

    .. _ClusterRoleBinding:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#clusterrolebinding-v1-rbac-authorization-k8s-io
    """

    obj_type = client.V1ClusterRoleBinding

    api_clients = {
        'preferred': client.RbacAuthorizationV1Api,
        'rbac.authorization.k8s.io/v1': client.RbacAuthorizationV1Api,
        'rbac.authorization.k8s.io/v1alpha1': client.RbacAuthorizationV1alpha1Api,
        'rbac.authorization.k8s.io/v1beta1': client.RbacAuthorizationV1beta1Api,
    }

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def create(self, namespace=None):
        """Create the ClusterRoleBinding under the given namespace.

        Args:
            namespace (str): This argument is ignored for ClusterRoleBindings.
        """
        log.info('creating clusterrolebinding "%s" in namespace "%s"', self.name, self.namespace)  # noqa
        log.debug('clusterrolebinding: %s', self.obj)

        self.obj = self.api_client.create_cluster_role_binding(
            body=self.obj,
        )

    def delete(self, options):
        """Delete the ClusterRoleBinding.

        This method expects the ClusterRoleBinding to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will need
        to be set manually.

        Args:
             options (client.V1DeleteOptions): Options for ClusterRoleBinding deletion.

        Returns:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting clusterrolebinding "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('clusterrolebinding: %s', self.obj)

        return self.api_client.delete_cluster_role_binding(
            name=self.name,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes ClusterRoleBinding resource."""
        self.obj = self.api_client.read_cluster_role_binding(
            name=self.name
        )

    def is_ready(self):
        """Check if the ClusterRoleBinding is in the ready state.

        ClusterRoleBindings do not have a "status" field to check, so we
        will measure their readiness status by whether or not they exist
        on the cluster.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        try:
            self.refresh()
        except:  # noqa
            return False
        else:
            return True



"""Setup and packaging for kubetest."""

import os

from codecs import open  # for consistent encoding
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))

# Load the package's __init__.py file as a dictionary.
pkg = {}
with open(os.path.join(here, 'kubetest', '__init__.py'), 'r', 'utf-8') as f:
    exec(f.read(), pkg)

# Load the README
readme = ''
if os.path.exists('README.md'):
    with open('README.md', 'r', 'utf-8') as f:
        readme = f.read()

setup(
    name=pkg['__title__'],
    version=pkg['__version__'],
    description=pkg['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    url=pkg['__url__'],
    author=pkg['__author__'],
    author_email=pkg['__author_email__'],
    license=pkg['__license__'],
    packages=find_packages(),
    python_requires='>=3.6',
    package_data={
        '': ['LICENSE'],
    },
    install_requires=[
        'kubernetes',
        'pyyaml>=4.2b1',
        'pytest',
        'requests>=2.21.0',  # used by 'kubernetes'
    ],
    zip_safe=False,
    classifiers=[
        'Environment :: Plugins',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    # this make a plugin available to pytest
    # https://docs.pytest.org/en/latest/writing_plugins.html#making-your-plugin-installable-by-others
    entry_points={
        'pytest11': [
            'kubetest = kubetest.plugin'
        ]
    }
)



"""Kubetest wrapper for the Kubernetes ``Endpoint`` API Object."""

import logging

from kubernetes import client

from .api_object import ApiObject

log = logging.getLogger('kubetest')


class Endpoints(ApiObject):
    """Kubetest wrapper around a Kubernetes `Endpoints`_ API Object.

    The actual ``kubernetes.client.V1Endpoints`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `Endpoints`_.

    .. _Endpoints:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#endpoints-v1-core
    """

    obj_type = client.V1Endpoints

    api_clients = {
        'preferred': client.CoreV1Api,
        'v1': client.CoreV1Api,
    }

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def create(self, namespace=None):
        """Create the Endpoints under the given namespace.

        Args:
            namespace (str): The namespace to create the Endpoints under.
        """
        if namespace is None:
            namespace = self.namespace

        log.info('creating endpoints "%s" in namespace "%s"', self.name, self.namespace)
        log.debug('endpoints: %s', self.obj)

        self.obj = self.api_client.create_namespaced_endpoints(
            namespace=namespace,
            body=self.obj,
        )

    def delete(self, options):
        """Delete the Endpoints.

        This method expects the Endpoints to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will need
        to be set manually.

        Args:
            options (client.V1DeleteOptions): Options for Endpoint deletion.

        Returns:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting endpoints "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('endpoints: %s', self.obj)

        return self.api_client.delete_namespaced_endpoints(
            name=self.name,
            namespace=self.namespace,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes Endpoints resource."""
        self.obj = self.api_client.read_namespaced_endpoints(
            name=self.name,
            namespace=self.namespace,
        )

    def is_ready(self):
        """Check if the Endpoints are in the ready state.

        The readiness of an Endpoint is determined by whether all of its
        subsets have no addresses in the "not ready" state.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        self.refresh()

        # check if any of the endpoint subsets has and addresses
        # in the not ready state.
        for subset in self.obj.subsets:
            if subset.not_ready_addresses > 0:
                return False
        return True



"""Kubetest wrapper for the Kubernetes ``Deployment`` API Object."""

import logging
import uuid

from kubernetes import client

from kubetest.utils import selector_string

from .api_object import ApiObject
from .pod import Pod

log = logging.getLogger('kubetest')


class Deployment(ApiObject):
    """Kubetest wrapper around a Kubernetes `Deployment`_ API Object.

    The actual ``kubernetes.client.V1Deployment`` instance that this
    wraps can be accessed via the ``obj`` instance member.

    This wrapper provides some convenient functionality around the
    API Object and provides some state management for the `Deployment`_.

    .. _Deployment:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.10/#deployment-v1-apps
    """

    obj_type = client.V1Deployment

    api_clients = {
        'preferred': client.AppsV1Api,
        'apps/v1': client.AppsV1Api,
        'apps/v1beta1': client.AppsV1beta1Api,
        'apps/v1beta2': client.AppsV1beta2Api,
    }

    def __init__(self, *args, **kwargs):
        super(Deployment, self).__init__(*args, **kwargs)
        self._add_kubetest_labels()

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()

    def _add_kubetest_labels(self):
        """Add a kubetest label to the Deployment object.

        This allows kubetest to more easily and reliably search for and aggregate
        API objects, such as getting the Pods for a Deployment.

        The kubetest label key is "kubetest/<obj kind>" where the obj kind is
        the lower-cased kind of the obj.
        """
        self.klabel_key = 'kubetest/deployment'
        if self.obj.metadata.labels:
            self.klabel_uid = self.obj.metadata.labels.get(self.klabel_key, None)
        else:
            self.klabel_uid = None
        if not self.klabel_uid:
            self.klabel_uid = str(uuid.uuid4())

        # fixme: it would be nice to clean up this label setting logic a bit
        #   and possibly abstract it out to something more generalized, but
        #   that is difficult to do given the differences in object attributes

        # Set the base metadata label
        if self.obj.metadata is None:
            self.obj.metadata = client.V1ObjectMeta()

        if self.obj.metadata.labels is None:
            self.obj.metadata.labels = {}

        if self.klabel_key not in self.obj.metadata.labels:
            self.obj.metadata.labels[self.klabel_key] = self.klabel_uid

        # If no spec is set, there is nothing to set additional labels on
        if self.obj.spec is None:
            log.warning('deployment spec not set - cannot set kubetest label')
            return

        # Set the selector label
        if self.obj.spec.selector is None:
            self.obj.spec.selector = client.V1LabelSelector()

        if self.obj.spec.selector.match_labels is None:
            self.obj.spec.selector.match_labels = {}

        if self.klabel_key not in self.obj.spec.selector.match_labels:
            self.obj.spec.selector.match_labels[self.klabel_key] = self.klabel_uid

        # Set the template label
        if self.obj.spec.template is None:
            self.obj.spec.template = client.V1PodTemplateSpec()

        if self.obj.spec.template.metadata is None:
            self.obj.spec.template.metadata = client.V1ObjectMeta(labels={})

        if self.klabel_key not in self.obj.spec.template.metadata.labels:
            self.obj.spec.template.metadata.labels[self.klabel_key] = self.klabel_uid

    def create(self, namespace=None):
        """Create the Deployment under the given namespace.

        Args:
            namespace (str): The namespace to create the Deployment under.
                If the Deployment was loaded via the kubetest client, the
                namespace will already be set, so it is not needed here.
                Otherwise, the namespace will need to be provided.
        """
        if namespace is None:
            namespace = self.namespace

        log.info('creating deployment "%s" in namespace "%s"', self.name, self.namespace)
        log.debug('deployment: %s', self.obj)

        self.obj = self.api_client.create_namespaced_deployment(
            namespace=namespace,
            body=self.obj,
        )

    def delete(self, options):
        """Delete the Deployment.

        This method expects the Deployment to have been loaded or otherwise
        assigned a namespace already. If it has not, the namespace will need
        to be set manually.

        Args:
            options (client.V1DeleteOptions): Options for Deployment deletion.

        Returns:
            client.V1Status: The status of the delete operation.
        """
        if options is None:
            options = client.V1DeleteOptions()

        log.info('deleting deployment "%s"', self.name)
        log.debug('delete options: %s', options)
        log.debug('deployment: %s', self.obj)

        return self.api_client.delete_namespaced_deployment(
            name=self.name,
            namespace=self.namespace,
            body=options,
        )

    def refresh(self):
        """Refresh the underlying Kubernetes Deployment resource."""
        self.obj = self.api_client.read_namespaced_deployment_status(
            name=self.name,
            namespace=self.namespace,
        )

    def is_ready(self):
        """Check if the Deployment is in the ready state.

        Returns:
            bool: True if in the ready state; False otherwise.
        """
        self.refresh()

        # if there is no status, the deployment is definitely not ready
        status = self.obj.status
        if status is None:
            return False

        # check the status for the number of total replicas and compare
        # it to the number of ready replicas. if the numbers are
        # equal, the deployment is ready; otherwise it is not ready.
        # TODO (etd) - we may want some logging in here eventually
        total = status.replicas
        ready = status.ready_replicas

        if total is None:
            return False

        return total == ready

    def status(self):
        """Get the status of the Deployment.

        Returns:
            client.V1DeploymentStatus: The status of the Deployment.
        """
        log.info('checking status of deployment "%s"', self.name)
        # first, refresh the deployment state to ensure the latest status
        self.refresh()

        # return the status from the deployment
        return self.obj.status

    def get_pods(self):
        """Get the pods for the Deployment.

        Returns:
            list[Pod]: A list of pods that belong to the deployment.
        """
        log.info('getting pods for deployment "%s"', self.name)

        pods = client.CoreV1Api().list_namespaced_pod(
            namespace=self.namespace,
            label_selector=selector_string({self.klabel_key: self.klabel_uid})
        )

        pods = [Pod(p) for p in pods.items]
        log.debug('pods: %s', pods)
        return pods
